# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
from typing import Mapping, Optional, TypeVar

from gt4py.eve import NodeTranslator, PreserveLocationVisitor
from gt4py.next.iterator import ir
from gt4py.next.iterator.ir_utils import ir_makers as im, misc as ir_misc
from gt4py.next.iterator.ir_utils.common_pattern_matcher import is_applied_lift
from gt4py.next.iterator.transforms.remap_symbols import RemapSymbolRefs, RenameSymbols
from gt4py.next.iterator.transforms.symbol_ref_utils import CountSymbolRefs
from gt4py.next.iterator.type_system import inference as itir_inference


# TODO(tehrengruber): Reduce complexity of the function by removing the different options here
#  and introduce a generic predicate argument for the `eligible_params` instead.
def inline_lambda(  # see todo above
    node: ir.FunCall,
    opcount_preserving=False,
    force_inline_lift_args=False,
    force_inline_trivial_lift_args=False,
    force_inline_lambda_args=False,
    eligible_params: Optional[list[bool]] = None,
):
    assert isinstance(node.fun, ir.Lambda)
    eligible_params = eligible_params or [True] * len(node.fun.params)

    assert len(eligible_params) == len(node.fun.params) == len(node.args)

    if opcount_preserving:
        ref_counts = CountSymbolRefs.apply(node.fun.expr, [p.id for p in node.fun.params])

        for i, param in enumerate(node.fun.params):
            # TODO(tehrengruber): allow inlining more complicated zero-op expressions like ignore_shift(...)(it_sym)
            if ref_counts[param.id] > 1 and not isinstance(
                node.args[i], (ir.SymRef, ir.Literal, ir.OffsetLiteral)
            ):
                eligible_params[i] = False

    # inline lifts, i.e. `lift(λ(...) → ...)(...)`
    if force_inline_lift_args:
        for i, arg in enumerate(node.args):
            if is_applied_lift(arg):
                eligible_params[i] = True

    # inline trivial lifts, i.e. `lift(λ() → 1)()`
    if force_inline_trivial_lift_args:
        for i, arg in enumerate(node.args):
            if is_applied_lift(arg) and len(arg.args) == 0:
                eligible_params[i] = True

    # inline lambdas passed as arguments
    if force_inline_lambda_args:
        for i, arg in enumerate(node.args):
            if isinstance(arg, ir.Lambda):
                eligible_params[i] = True

    if node.fun.params and not any(eligible_params):
        return node

    refs: set[str] = set().union(
        *(
            arg.pre_walk_values().if_isinstance(ir.SymRef).getattr("id").to_set()
            for arg, eligible in zip(node.args, eligible_params)
            if eligible
        )
    )
    syms: set[str] = node.fun.pre_walk_values().if_isinstance(ir.Sym).getattr("id").to_set()
    clashes = refs & syms
    fun = node.fun
    if clashes:
        # TODO(tehrengruber): find a better way of generating new symbols in `name_map` that don't collide with each other. E.g. this must still work:
        # (lambda arg, arg_: (lambda arg_: ...)(arg))(a, b)  # noqa: ERA001 [commented-out-code]
        name_map: dict[str, str] = {}

        for sym in clashes:
            name_map[sym] = ir_misc.unique_symbol(sym, refs | syms | {*name_map.values()})

        # Let's rename the symbols (including params) of the function.
        # If we would like to preserve the original param names, we could alternatively
        # rename the eligible symrefs in `args`.
        fun = RenameSymbols().visit(fun, name_map=name_map)

    symbol_map = {
        param.id: arg
        for param, arg, eligible in zip(fun.params, node.args, eligible_params)
        if eligible
    }
    new_expr = RemapSymbolRefs().visit(fun.expr, symbol_map=symbol_map)

    if all(eligible_params):
        new_expr.location = node.location
    else:
        new_expr = ir.FunCall(
            fun=ir.Lambda(
                params=[
                    param for param, eligible in zip(fun.params, eligible_params) if not eligible
                ],
                expr=new_expr,
            ),
            args=[arg for arg, eligible in zip(node.args, eligible_params) if not eligible],
            location=node.location,
        )
    for attr in ("type", "recorded_shifts", "domain"):
        if hasattr(node.annex, attr):
            setattr(new_expr.annex, attr, getattr(node.annex, attr))
    itir_inference.copy_type(from_=node, to=new_expr, allow_untyped=True)
    return new_expr


T = TypeVar("T", bound=ir.Expr)


def rename_symbols(node: T, rename_map: Mapping[str, str | ir.SymRef]) -> T:
    """
    Given a node and a mapping from old symbol names to new symbol names, rename symbols.

    >>> str(rename_symbols(im.plus("old_name", 1), {"old_name": "new_name"}))
    'new_name + 1'
    """
    return inline_lambda(im.let(*rename_map.items())(node))


@dataclasses.dataclass
class InlineLambdas(PreserveLocationVisitor, NodeTranslator):
    """
    Inline lambda calls by substituting every argument by its value.

    Note: This pass preserves, but doesn't use the `type` `recorded_shifts`, `domain` annex.
    """

    PRESERVED_ANNEX_ATTRS = ("type", "recorded_shifts", "domain")

    opcount_preserving: bool

    force_inline_lambda_args: bool

    force_inline_lift_args: bool

    force_inline_trivial_lift_args: bool

    @classmethod
    def apply(
        cls,
        node: ir.Node,
        opcount_preserving=False,
        force_inline_lambda_args=False,
        force_inline_lift_args=False,
        force_inline_trivial_lift_args=False,
    ):
        """
        Inline lambda calls by substituting every argument by its value.

        Examples:
            `(λ(x) → x)(y)` to `y`
            `(λ(x) → x)(y+y)` to `y+y`
            `(λ(x) → x+x)(y+y)` to `y+y+y+y` if not opcount_preserving
            `(λ(x) → x+x)(y+y)` stays as is if opcount_preserving

        Arguments:
            node: The function call node to inline into.
            opcount_preserving: Preserve the number of operations, i.e. only
                inline lambda call if the resulting call has the same number of
                operations.
            force_inline_lambda_args: Inline all arguments that are lambda calls, i.e.
                `(λ(p) → p(a, a))(λ(x, y) → x+y)`
            force_inline_lift_args: Inline all arguments that are applied lifts, i.e.
                `lift(λ(...) → ...)(...)`.
            force_inline_trivial_lift_args: Inline all arguments that are trivial
                applied lifts, e.g. `lift(λ() → 1)()`.

        """
        return cls(
            opcount_preserving=opcount_preserving,
            force_inline_lambda_args=force_inline_lambda_args,
            force_inline_lift_args=force_inline_lift_args,
            force_inline_trivial_lift_args=force_inline_trivial_lift_args,
        ).visit(node)

    def visit_FunCall(self, node: ir.FunCall):
        node = self.generic_visit(node)
        if isinstance(node.fun, ir.Lambda):
            return inline_lambda(
                node,
                opcount_preserving=self.opcount_preserving,
                force_inline_lambda_args=self.force_inline_lambda_args,
                force_inline_lift_args=self.force_inline_lift_args,
                force_inline_trivial_lift_args=self.force_inline_trivial_lift_args,
            )

        return node



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import ast
import copy
from collections.abc import Iterator


class NodeYielder(ast.NodeTransformer):
    @classmethod
    def apply(cls, node: ast.AST) -> ast.AST:
        result = list(cls().visit(node))
        if len(result) != 1:
            raise ValueError("AST was split or lost during the pass, use '.visit()' instead.")
        return result[0]

    def visit(self, node: ast.AST) -> Iterator[ast.AST]:
        result = super().visit(node)
        if isinstance(result, ast.AST):
            yield result
        else:
            yield from result

    def generic_visit(self, node: ast.AST) -> Iterator[ast.AST]:  # type: ignore[override]
        """Override generic visit to deal with generators."""
        for field, old_value in ast.iter_fields(node):
            if isinstance(old_value, list):
                new_values = [i for j in old_value for i in self.visit(j)]
                old_value[:] = new_values
            elif isinstance(old_value, ast.AST):
                new_node, *_ = list(self.visit(old_value)) or (None,)
                if new_node is None:
                    delattr(node, field)
                else:
                    setattr(node, field, new_node)
        yield node


class SingleAssignTargetPass(NodeYielder):
    """
    Split multi target assignments.

    Requires AST in SSA form (see ``SingleStaticAssignPass``) to yield correct results.

    Example
    -------
    >>> import ast, inspect
    >>>
    >>> def foo():
    ...     a = b = 1
    ...     return a, b
    >>>
    >>> print(ast.unparse(SingleAssignTargetPass.apply(ast.parse(inspect.getsource(foo)))))
    def foo():
        __sat_tmp0 = 1
        a = __sat_tmp0
        b = __sat_tmp0
        return (a, b)

    """

    unique_name_id: int = 0

    def _unique_symbol_name(self) -> ast.Name:
        name = ast.Name(id=f"__sat_tmp{self.unique_name_id}", ctx=ast.Store())
        self.unique_name_id += 1
        return name

    def visit_Assign(self, node: ast.Assign) -> Iterator[ast.Assign]:
        # ignore regular assignments
        if len(node.targets) == 1:
            yield node
            return

        synthetic_target = self._unique_symbol_name()
        ast.copy_location(synthetic_target, node)
        synthetic_assign = copy.copy(node)
        synthetic_assign.targets = [synthetic_target]
        yield synthetic_assign

        for target in node.targets:
            new_assign = copy.copy(node)
            new_assign.targets = [target]
            new_assign.value = ast.Name(id=synthetic_target.id, ctx=ast.Load())
            yield new_assign



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from typing import Protocol, TypeVar

from gt4py.next.otf import languages, stages, workflow


SrcL = TypeVar("SrcL", bound=languages.LanguageTag)
TgtL = TypeVar("TgtL", bound=languages.LanguageTag)
LS = TypeVar("LS", bound=languages.LanguageSettings)
SrcL_co = TypeVar("SrcL_co", bound=languages.LanguageTag, covariant=True)
TgtL_co = TypeVar("TgtL_co", bound=languages.LanguageTag, covariant=True)
LS_co = TypeVar("LS_co", bound=languages.LanguageSettings, covariant=True)


class TranslationStep(
    workflow.ReplaceEnabledWorkflowMixin[stages.CompilableProgram, stages.ProgramSource[SrcL, LS]],
    Protocol[SrcL, LS],
):
    """Translate a GT4Py program to source code (ProgramCall -> ProgramSource)."""

    ...


class BindingStep(Protocol[SrcL, LS, TgtL]):
    """
    Generate Bindings for program source and package both together (ProgramSource -> CompilableSource).

    In the special cases where bindings are not required, such a step could also simply construct
    a ``CompilableSource`` from the ``ProgramSource`` with bindings set to ``None``.
    """

    def __call__(
        self, program_source: stages.ProgramSource[SrcL, LS]
    ) -> stages.CompilableSource[SrcL, LS, TgtL]: ...


class CompilationStep(
    workflow.Workflow[stages.CompilableSource[SrcL, LS, TgtL], stages.CompiledProgram],
    Protocol[SrcL, LS, TgtL],
):
    """Compile program source code and bindings into a python callable (CompilableSource -> CompiledProgram)."""

    def __call__(
        self, source: stages.CompilableSource[SrcL, LS, TgtL]
    ) -> stages.CompiledProgram: ...



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import collections.abc
import copy
import dataclasses
import functools

from gt4py import eve
from gt4py.eve import concepts
from gt4py.eve.extended_typing import Any, Callable, Optional, TypeVar, Union
from gt4py.next import common
from gt4py.next.iterator import builtins, ir as itir
from gt4py.next.iterator.ir_utils.common_pattern_matcher import is_applied_as_fieldop, is_call_to
from gt4py.next.iterator.transforms import symbol_ref_utils, trace_shifts
from gt4py.next.iterator.type_system import type_specifications as it_ts, type_synthesizer
from gt4py.next.type_system import type_info, type_specifications as ts
from gt4py.next.type_system.type_info import primitive_constituents


def _is_representable_as_int(s: int | str) -> bool:
    try:
        int(s)
        return True
    except ValueError:
        return False


def _set_node_type(node: itir.Node, type_: ts.TypeSpec) -> None:
    if node.type:
        assert type_info.is_compatible_type(node.type, type_), (
            f"Node {node!s} already has a type {node.type} which differs from {type_}."
        )
    # Also populate the type of the parameters of a lambda. That way the one can access the type
    # of a parameter by a lookup in the symbol table. As long as `_set_node_type` is used
    # exclusively, the information stays consistent with the types stored in the `FunctionType`
    # of the lambda itself.
    if isinstance(node, itir.Lambda):
        assert isinstance(type_, ts.FunctionType)
        for param, param_type in zip(node.params, type_.pos_only_args):
            _set_node_type(param, param_type)
    node.type = type_


def copy_type(from_: itir.Node, to: itir.Node, allow_untyped: bool = False) -> None:
    """
    Copy type from one node to another.

    This function mainly exists for readability reasons.
    """
    assert allow_untyped is not None or isinstance(from_.type, ts.TypeSpec)
    if from_.type is None:
        assert allow_untyped
        return
    _set_node_type(to, from_.type)


def on_inferred(callback: Callable, *args: Union[ts.TypeSpec, ObservableTypeSynthesizer]) -> None:
    """
    Execute `callback` as soon as all `args` have a type.
    """
    ready_args = [False] * len(args)
    inferred_args = [None] * len(args)

    def mark_ready(i, type_):
        ready_args[i] = True
        inferred_args[i] = type_
        if all(ready_args):
            callback(*inferred_args)

    for i, arg in enumerate(args):
        if isinstance(arg, ObservableTypeSynthesizer):
            arg.on_type_ready(functools.partial(mark_ready, i))
        else:
            assert isinstance(arg, ts.TypeSpec)
            mark_ready(i, arg)


@dataclasses.dataclass
class ObservableTypeSynthesizer(type_synthesizer.TypeSynthesizer):
    """
    This class wraps a type synthesizer to handle typing of nodes representing functions.

    The type inference algorithm represents functions as type synthesizer, i.e. regular
    callables that given a set of arguments compute / deduce the return type. The return type of
    functions, let it be a builtin like ``itir.plus`` or a user defined lambda function, is only
    defined when all its arguments are typed.

    Let's start with a small example to exemplify this. The power function has a rather simple
    type synthesizer, where the output type is simply the type of the base.

    >>> def power(base: ts.ScalarType, exponent: ts.ScalarType) -> ts.ScalarType:
    ...     return base
    >>> float_type = ts.ScalarType(kind=ts.ScalarKind.FLOAT64)
    >>> int_type = ts.ScalarType(kind=ts.ScalarKind.INT64)
    >>> power(float_type, int_type)
    ScalarType(kind=<ScalarKind.FLOAT64: 11>, shape=None)

    Now, consider a simple lambda function that squares its argument using the power builtin. A
    type synthesizer for this function is simple to formulate, but merely gives us the return
    type of the function.

    >>> from gt4py.next.iterator.ir_utils import ir_makers as im
    >>> square_func = im.lambda_("base")(im.call("power")("base", 2))
    >>> square_func_type_synthesizer = type_synthesizer.type_synthesizer(
    ...     lambda base: power(base, int_type)
    ... )
    >>> square_func_type_synthesizer(float_type, offset_provider_type={})
    ScalarType(kind=<ScalarKind.FLOAT64: 11>, shape=None)

    Note that without a corresponding call the function itself can not be fully typed and as such
    the type inference algorithm has to defer typing until then. This task is handled transparently
    (in the sense that an ``ObservableTypeSynthesizer`` is a type synthesizer again) by this
    class. Given a type synthesizer and a node we obtain a new type synthesizer that when
    evaluated stores the type of the function in the node.

    >>> o_type_synthesizer = ObservableTypeSynthesizer(
    ...     type_synthesizer=square_func_type_synthesizer,
    ...     node=square_func,
    ...     store_inferred_type_in_node=True,
    ... )
    >>> o_type_synthesizer(float_type, offset_provider_type={})
    ScalarType(kind=<ScalarKind.FLOAT64: 11>, shape=None)
    >>> square_func.type == ts.FunctionType(
    ...     pos_only_args=[float_type], pos_or_kw_args={}, kw_only_args={}, returns=float_type
    ... )
    True

    Note that this is a simple example where the type of the arguments and the return value is
    available when the function is called. In order to support higher-order functions, where
    arguments or return value are functions itself (i.e. passed as type rules) this class provides
    additional functionality for multiple typing rules to notify each other about a type being
    ready.
    """

    #: node that has this type
    node: Optional[itir.Node] = None
    #: list of references to this function
    aliases: list[itir.SymRef] = dataclasses.field(default_factory=list)
    #: list of callbacks executed as soon as the type is ready
    callbacks: list[Callable[[ts.TypeSpec], None]] = dataclasses.field(default_factory=list)
    #: the inferred type when ready and None until then
    inferred_type: Optional[ts.FunctionType] = None
    #: whether to store the type in the node or not
    store_inferred_type_in_node: bool = False

    def infer_type(
        self, return_type: ts.DataType | ts.DeferredType, *args: ts.DataType | ts.DeferredType
    ) -> ts.FunctionType:
        return ts.FunctionType(
            pos_only_args=list(args), pos_or_kw_args={}, kw_only_args={}, returns=return_type
        )

    def _infer_type_listener(self, return_type: ts.TypeSpec, *args: ts.TypeSpec) -> None:
        self.inferred_type = self.infer_type(return_type, *args)  # type: ignore[arg-type]  # ensured by assert above

        # if the type has been fully inferred, notify all `ObservableTypeSynthesizer`s that depend on it.
        for cb in self.callbacks:
            cb(self.inferred_type)

        if self.store_inferred_type_in_node:
            assert self.node
            _set_node_type(self.node, self.inferred_type)
            self.node.type = self.inferred_type
            for alias in self.aliases:
                _set_node_type(alias, self.inferred_type)

    def on_type_ready(self, cb: Callable[[ts.TypeSpec], None]) -> None:
        if self.inferred_type:
            # type has already been inferred, just call the callback
            cb(self.inferred_type)
        else:
            self.callbacks.append(cb)

    def __call__(
        self,
        *args: type_synthesizer.TypeOrTypeSynthesizer,
        offset_provider_type: common.OffsetProviderType,
        **kwargs,
    ) -> Union[ts.TypeSpec, ObservableTypeSynthesizer]:
        assert all(isinstance(arg, (ts.TypeSpec, ObservableTypeSynthesizer)) for arg in args), (
            "ObservableTypeSynthesizer can only be used with arguments that are TypeSpec or ObservableTypeSynthesizer"
        )

        return_type_or_synthesizer = self.type_synthesizer(
            *args, offset_provider_type=offset_provider_type, **kwargs
        )

        # return type is a typing rule by itself
        if isinstance(return_type_or_synthesizer, type_synthesizer.TypeSynthesizer):
            return_type_or_synthesizer = ObservableTypeSynthesizer(
                node=None,  # node will be set by caller
                type_synthesizer=return_type_or_synthesizer,
                store_inferred_type_in_node=True,
            )

        assert isinstance(return_type_or_synthesizer, (ts.TypeSpec, ObservableTypeSynthesizer))

        # delay storing the type until the return type and all arguments are inferred
        on_inferred(self._infer_type_listener, return_type_or_synthesizer, *args)  # type: ignore[arg-type] # ensured by assert above

        return return_type_or_synthesizer


def _get_dimensions_from_offset_provider(
    offset_provider_type: common.OffsetProviderType,
) -> dict[str, common.Dimension]:
    dimensions: dict[str, common.Dimension] = {}
    for offset_name, provider in offset_provider_type.items():
        dimensions[offset_name] = common.Dimension(
            value=offset_name, kind=common.DimensionKind.LOCAL
        )
        if isinstance(provider, common.Dimension):
            dimensions[provider.value] = provider
        elif isinstance(provider, common.NeighborConnectivityType):
            dimensions[provider.source_dim.value] = provider.source_dim
            dimensions[provider.codomain.value] = provider.codomain
    return dimensions


def _get_dimensions_from_types(types) -> dict[str, common.Dimension]:
    def _get_dimensions(obj: Any):
        if isinstance(obj, common.Dimension):
            yield obj
        elif isinstance(obj, ts.TypeSpec):
            for field in obj.__datamodel_fields__.keys():
                yield from _get_dimensions(getattr(obj, field))
        elif isinstance(obj, collections.abc.Mapping):
            for el in obj.values():
                yield from _get_dimensions(el)
        elif isinstance(obj, collections.abc.Iterable) and not isinstance(obj, str):
            for el in obj:
                yield from _get_dimensions(el)

    return {dim.value: dim for dim in _get_dimensions(types)}


def _type_synthesizer_from_function_type(fun_type: ts.FunctionType):
    def type_synthesizer(*args, **kwargs):
        assert type_info.accepts_args(fun_type, with_args=list(args), with_kwargs=kwargs)
        return fun_type.returns

    return ObservableTypeSynthesizer(
        type_synthesizer=type_synthesizer, store_inferred_type_in_node=False
    )


class SanitizeTypes(eve.NodeTranslator, eve.VisitorWithSymbolTableTrait):
    PRESERVED_ANNEX_ATTRS = ("domain",)

    def visit_Node(self, node: itir.Node, *, symtable: dict[str, itir.Node]) -> itir.Node:
        node = self.generic_visit(node)
        # We only want to sanitize types that have been inferred previously such that we don't run
        # into errors because a node has been reused in a pass, but has changed type. Undeclared
        # symbols however only occur when visiting a subtree (e.g. in testing). Their types
        # can be injected by populating their type attribute, which we want to preserve here.
        is_undeclared_symbol = isinstance(node, itir.SymRef) and node.id not in symtable
        if not is_undeclared_symbol and not isinstance(node, (itir.Literal, itir.Sym)):
            node.type = None
        return node


T = TypeVar("T", bound=itir.Node)

_INITIAL_CONTEXT = {
    name: ObservableTypeSynthesizer(
        type_synthesizer=type_synthesizer.builtin_type_synthesizers[name],
        # builtin functions are polymorphic
        store_inferred_type_in_node=False,
    )
    for name in type_synthesizer.builtin_type_synthesizers.keys()
}


@dataclasses.dataclass
class ITIRTypeInference(eve.NodeTranslator):
    """
    ITIR type inference algorithm.

    See :method:ITIRTypeInference.apply for more details.
    """

    PRESERVED_ANNEX_ATTRS = ("domain",)

    offset_provider_type: Optional[common.OffsetProviderType]
    #: Allow sym refs to symbols that have not been declared. Mostly used in testing.
    allow_undeclared_symbols: bool
    #: Reinference-mode skipping already typed nodes.
    reinfer: bool

    @classmethod
    def apply(
        cls,
        node: T,
        *,
        offset_provider_type: common.OffsetProviderType,
        inplace: bool = False,
        allow_undeclared_symbols: bool = False,
    ) -> T:
        """
        Infer the type of ``node`` and its sub-nodes.

        Arguments:
            node: The :class:`itir.Node` to infer the types of.

        Keyword Arguments:
            offset_provider_type: Offset provider dictionary.
            inplace: Write types directly to the given ``node`` instead of returning a copy.
            allow_undeclared_symbols: Allow references to symbols that don't have a corresponding
              declaration. This is useful for testing or inference on partially inferred sub-nodes.

        Preconditions:

        All parameters in :class:`itir.Program` must have a type
        defined, as they are the starting point for type propagation.

        Design decisions:
        - Lambda functions are monomorphic
        Builtin functions like ``plus`` are by design polymorphic and only their argument and return
        types are of importance in transformations. Lambda functions on the contrary also have a
        body on which we would like to run transformations. By choosing them to be monomorphic all
        types in the body can be inferred to a concrete type, making reasoning about them in
        transformations simple. Consequently, the following is invalid as `f` is called with
        arguments of different type
        ```
        let f = λ(a) → a+a
            in f(1)+f(1.)
        ```
        In case we want polymorphic lambda functions, i.e. generic functions in the frontend
        could be implemented that way, current consensus is to instead implement a transformation
        that duplicates the lambda function for each of the types it is called with
        ```
        let f_int = λ(a) → a+a, f_float = λ(a) → a+a
            in f_int(1)+f_float(1.)
        ```
        Note that this is not the only possible choice. Polymorphic lambda functions and a type
        inference algorithm that only infers the most generic type would allow us to run
        transformations without this duplication and reduce code size early. However, this would
        require careful planning and documentation on what information a transformation needs.

        Limitations:

        - The current position of (iterator) arguments to a lifted stencil is unknown
        Consider the following trivial stencil: ``λ(it) → deref(it)``. A priori we don't know
        what the current position of ``it`` is (inside the body of the lambda function), but only
        when we call the stencil with an actual iterator the position becomes known. Consequently,
        when we lift the stencil, the position of its iterator arguments is only known as soon as
        the iterator as returned by the applied lift is dereferenced. Deferring the inference
        of the current position for lifts has been decided to be too complicated as we don't need
        the information right now and is hence not implemented.

        - Iterators only ever reference values, not columns.
        The initial version of the ITIR used iterators of columns and vectorized operations between
        columns in order to express scans. This differentiation is not needed in our transformations
        and as such was not implemented here.
        """
        # TODO(tehrengruber): some of the transformations reuse nodes with type information that
        #  becomes invalid (e.g. the shift part of ``shift(...)(it)`` has a different type when used
        #  on a different iterator). For now we just delete all types in case we are working an
        #   parts of a program.
        node = SanitizeTypes().visit(node)

        if isinstance(node, itir.Program):
            assert all(isinstance(param.type, ts.DataType) for param in node.params), (
                "All parameters in 'itir.Program' must have a type "
                "defined, as they are the starting point for type propagation.",
            )

        instance = cls(
            offset_provider_type=offset_provider_type,
            allow_undeclared_symbols=allow_undeclared_symbols,
            reinfer=False,
        )
        if not inplace:
            node = copy.deepcopy(node)
        instance.visit(node, ctx=_INITIAL_CONTEXT)
        return node

    @classmethod
    def apply_reinfer(
        cls, node: T, *, offset_provider_type: Optional[common.OffsetProviderType] = None
    ) -> T:
        """
        Given a partially typed node infer the type of ``node`` and its sub-nodes.

        Contrary to the regular inference, this method does not descend into already typed sub-nodes
        and can be used as a lightweight way to restore type information during a pass.

        Note that this function alters the input node, which is usually desired, and more
        performant.

        Arguments:
            node: The :class:`itir.Node` to infer the types of.
            offset_provider_type: Offset provider dictionary.
        """
        if node.type:  # already inferred
            return node

        instance = cls(
            offset_provider_type=offset_provider_type, allow_undeclared_symbols=True, reinfer=True
        )
        instance.visit(node, ctx=_INITIAL_CONTEXT)
        return node

    def visit(self, node: concepts.RootNode, **kwargs: Any) -> Any:
        # we found a node that is typed, do not descend into children
        if self.reinfer and isinstance(node, itir.Node) and node.type:
            if isinstance(node.type, ts.FunctionType):
                return _type_synthesizer_from_function_type(node.type)
            return node.type

        result = super().visit(node, **kwargs)

        if isinstance(node, itir.Node):
            if isinstance(result, ts.TypeSpec):
                if node.type and not isinstance(node.type, ts.DeferredType):
                    assert type_info.is_compatible_type(node.type, result)
                node.type = result
            elif isinstance(result, ObservableTypeSynthesizer) or result is None:
                pass
            elif isinstance(result, type_synthesizer.TypeSynthesizer):
                # this case occurs either when a Lambda node is visited or TypeSynthesizer returns
                # another type synthesizer.
                return ObservableTypeSynthesizer(
                    node=node,
                    type_synthesizer=result,
                    store_inferred_type_in_node=True,
                )
            else:
                raise AssertionError(
                    f"Expected a 'TypeSpec', `TypeSynthesizer` or 'ObservableTypeSynthesizer', "
                    f"`but got {type(result).__name__}`"
                )
        return result

    def visit_Program(self, node: itir.Program, *, ctx) -> it_ts.ProgramType:
        params: dict[str, ts.DataType] = {}
        for param in node.params:
            assert isinstance(param.type, ts.DataType)
            params[param.id] = param.type
        decls: dict[str, ts.FieldType] = {}
        for fun_def in node.function_definitions:
            decls[fun_def.id] = self.visit(fun_def, ctx=ctx | params | decls)
        for decl_node in node.declarations:
            decls[decl_node.id] = self.visit(decl_node, ctx=ctx | params | decls)
        self.visit(node.body, ctx=ctx | params | decls)
        return it_ts.ProgramType(params=params)

    def visit_Temporary(self, node: itir.Temporary, *, ctx) -> ts.FieldType | ts.TupleType:
        domain = self.visit(node.domain, ctx=ctx)
        assert isinstance(domain, ts.DomainType)
        assert domain.dims != "unknown"
        assert node.dtype
        return type_info.apply_to_primitive_constituents(
            lambda dtype: ts.FieldType(dims=domain.dims, dtype=dtype),
            node.dtype,
        )

    def visit_IfStmt(self, node: itir.IfStmt, *, ctx) -> None:
        cond = self.visit(node.cond, ctx=ctx)
        assert cond == ts.ScalarType(kind=ts.ScalarKind.BOOL)
        self.visit(node.true_branch, ctx=ctx)
        self.visit(node.false_branch, ctx=ctx)

    def visit_SetAt(self, node: itir.SetAt, *, ctx) -> None:
        self.visit(node.expr, ctx=ctx)
        self.visit(node.domain, ctx=ctx)
        self.visit(node.target, ctx=ctx)
        assert node.target.type is not None and node.expr.type is not None
        for target_type, path in primitive_constituents(node.target.type, with_path_arg=True):
            # the target can have fewer elements than the expr in which case the output from the
            # expression is simply discarded.
            expr_type = functools.reduce(
                lambda tuple_type, i: tuple_type.types[i]  # type: ignore[attr-defined]  # format ensured by primitive_constituents
                # `ts.DeferredType` only occurs for scans returning a tuple
                if not isinstance(tuple_type, ts.DeferredType)
                else ts.DeferredType(constraint=None),
                path,
                node.expr.type,
            )
            assert isinstance(target_type, (ts.FieldType, ts.DeferredType))
            assert isinstance(expr_type, (ts.FieldType, ts.DeferredType))
            if isinstance(target_type, ts.FieldType) and isinstance(expr_type, ts.FieldType):
                assert expr_type.dims == target_type.dims
                assert target_type.dtype == expr_type.dtype

    def visit_AxisLiteral(self, node: itir.AxisLiteral, **kwargs) -> ts.DimensionType:
        return ts.DimensionType(dim=common.Dimension(value=node.value, kind=node.kind))

    # TODO: revisit what we want to do with OffsetLiterals as we already have an Offset type in
    #  the frontend.
    def visit_OffsetLiteral(self, node: itir.OffsetLiteral, **kwargs) -> it_ts.OffsetLiteralType:
        if _is_representable_as_int(node.value):
            return it_ts.OffsetLiteralType(
                value=ts.ScalarType(
                    kind=getattr(ts.ScalarKind, builtins.INTEGER_INDEX_BUILTIN.upper())
                )
            )
        else:
            assert isinstance(node.value, str)
            return it_ts.OffsetLiteralType(value=node.value)

    def visit_Literal(self, node: itir.Literal, **kwargs) -> ts.ScalarType:
        assert isinstance(node.type, ts.ScalarType)
        return node.type

    def visit_InfinityLiteral(self, node: itir.InfinityLiteral, **kwargs) -> ts.ScalarType:
        return ts.ScalarType(kind=ts.ScalarKind.INT32)

    def visit_NegInfinityLiteral(self, node: itir.InfinityLiteral, **kwargs) -> ts.ScalarType:
        return ts.ScalarType(kind=ts.ScalarKind.INT32)

    def visit_SymRef(
        self, node: itir.SymRef, *, ctx: dict[str, ts.TypeSpec]
    ) -> ts.TypeSpec | type_synthesizer.TypeSynthesizer:
        # for testing, it is useful to be able to use types without a declaration
        if self.allow_undeclared_symbols and node.id not in ctx:
            # type has been stored in the node itself
            if node.type:
                if isinstance(node.type, ts.FunctionType):
                    return _type_synthesizer_from_function_type(node.type)
                return node.type
            return ts.DeferredType(constraint=None)
        assert node.id in ctx
        result = ctx[node.id]
        if isinstance(result, ObservableTypeSynthesizer):
            result.aliases.append(node)
        return result

    def visit_Lambda(
        self, node: itir.Lambda | itir.FunctionDefinition, *, ctx: dict[str, ts.TypeSpec]
    ) -> type_synthesizer.TypeSynthesizer:
        @type_synthesizer.type_synthesizer(cache=True)
        def fun(*args):
            return self.visit(
                node.expr, ctx=ctx | {p.id: a for p, a in zip(node.params, args, strict=True)}
            )

        return fun

    visit_FunctionDefinition = visit_Lambda

    def visit_FunCall(
        self, node: itir.FunCall, *, ctx: dict[str, ts.TypeSpec]
    ) -> ts.TypeSpec | type_synthesizer.TypeSynthesizer:
        # grammar builtins
        if is_call_to(node, "cast_"):
            value, type_constructor = node.args
            self.visit(value, ctx=ctx)  # ensure types in value are also inferred
            assert (
                isinstance(type_constructor, itir.SymRef)
                and type_constructor.id in builtins.TYPE_BUILTINS
            )
            return ts.ScalarType(kind=getattr(ts.ScalarKind, type_constructor.id.upper()))

        if is_call_to(node, "tuple_get"):
            index_literal, tuple_ = node.args
            self.visit(tuple_, ctx=ctx)  # ensure tuple is typed
            assert isinstance(index_literal, itir.Literal)
            index = int(index_literal.value)
            if isinstance(tuple_.type, ts.DeferredType):
                return ts.DeferredType(constraint=None)
            assert isinstance(tuple_.type, ts.TupleType)
            return tuple_.type.types[index]

        # Additional information beyond what is given by the argument types
        syntactic_info = {}
        if is_applied_as_fieldop(node):
            # Collect shifts and pass them as a kwarg to :func:`type_synthesizer.applied_as_fieldop`.
            # The node is available here, but not within the `TypeSynthesizer`, so we handle it here.
            stencil = node.fun.args[0]
            referenced_fun_names = symbol_ref_utils.collect_symbol_refs(stencil)
            syntactic_info["shift_sequences_per_param"] = (
                trace_shifts.trace_stencil(stencil, num_args=len(node.args))
                if not referenced_fun_names
                else None
            )

        fun = self.visit(node.fun, ctx=ctx)
        args = self.visit(node.args, ctx=ctx)
        result = fun(*args, **syntactic_info, offset_provider_type=self.offset_provider_type)

        if isinstance(result, ObservableTypeSynthesizer):
            assert not result.node
            result.node = node

        return result

    def visit_Node(self, node: itir.Node, **kwargs):
        raise NotImplementedError(f"No type rule for nodes of type '{type(node).__name__}'.")


infer = ITIRTypeInference.apply

reinfer = ITIRTypeInference.apply_reinfer



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import copy
import dataclasses
from typing import Any, Iterable, Optional, Sequence, Union

import dace
from dace import data as dace_data, subsets as dace_sbs, symbolic as dace_sym
from dace.sdfg import graph as dace_graph, nodes as dace_nodes

from gt4py.next.program_processors.runners.dace import transformations as gtx_transformations


@dataclasses.dataclass(frozen=True)
class EdgeConnectionSpec:
    """Describes an edge in an abstract way, that is kind of independent of the direction.

    It is always described in terms of a node, which is eader the `src` or the `dst`
    of the edge, in terms of the subset the edge read or writes and the edge itself.

    To construct `EdgeConnectionSpec` you can use the `describe_incoming_edges()`
    and `describe_outgoing_edges()` function.
    """

    node: dace_nodes.Node
    subset: dace_sbs.Subset
    edge: dace_graph.MultiConnectorEdge

    def __post_init__(self) -> None:
        assert self.subset is not None
        assert all((r[2] == 1) == True for r in self.subset)  # noqa: E712 [true-false-comparison]  # SymPy comparison

    def __hash__(self) -> int:
        return hash(self.edge)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, EdgeConnectionSpec):
            return self.edge.other.self
        return self.edge == other

    @property
    def is_incoming(self) -> bool:
        """Checks if the edge describes an incoming edge.

        An edge is incoming if `self.node` is the same as `self.edge.dst`.
        """
        return self.node is self.edge.dst

    @property
    def other_node(self) -> dace_nodes.Node:
        """Returns the node at the other side of the edge."""
        return self.edge.src if self.is_incoming else self.edge.dst

    @property
    def other_subset(self) -> Optional[dace_sbs.Subset]:
        """Returns the subset at the other end of the edge.

        It is important that there is no real connection with the `other_subset`
        attribute of `dace.Memlet`, the differences are subtitle but important.
        `self.subset` is always the subset the edge has at the node that it
        describes, while `self.other_subset` is "the other one".

        Note:
            While `self.subset` is never `None`, this property might be `None`.
        """
        return self.edge.data.src_subset if self.is_incoming else self.edge.data.dst_subset


def describe_edge(
    edge: dace_graph.MultiConnectorEdge,
    incoming_edge: bool,
    state: Optional[dace.SDFGState] = None,
) -> EdgeConnectionSpec:
    """Create a description for a single edge.

    Args:
        edge: The edge to describe.
        incoming_edge: Describe whether it is an incoming (`True`) or outgoing (`False`)
            edge, used to select the destination or source subset.
        state: If it is passed, the memlet subset will be extracted using `get_{src, dst}_subset()`.
            If `None`, the default, the `{src, dst}_subset` property will be used.
            Only pass `state` if the state is not inconsistent.
    """
    get_sbs = lambda e: e.data.dst_subset if incoming_edge else e.data.src_subset  # noqa: E731 [lambda-assignment]
    get_node = lambda e: e.dst if incoming_edge else e.src  # noqa: E731 [lambda-assignment]
    if state is not None:
        # To ensures that the `{src,dst}_subset` are properly set, run initialization.
        #  See [issue 1708](https://github.com/spcl/dace/issues/1703)
        edge.data.try_initialize(state.sdfg, state, edge)
    return EdgeConnectionSpec(
        node=get_node(edge),
        subset=get_sbs(edge),
        edge=edge,
    )


def describe_edges(
    state: dace.SDFGState,
    node: dace_nodes.Node,
    incoming_edges: bool,
) -> list[EdgeConnectionSpec]:
    """Generate the description of the edges.

    Function will either describes the incoming edges on node `node`, if
    `incoming_edges` is `True` or the out going edges of node `node` if
    `incoming_edges` is `False`.

    Args:
        state: The state in which `node` is located.
        node: The node whose edges should be described.
        incoming_edges: Describe the incoming (`True`) or out going (`False`) edges
            of `node`.
    """
    edges = state.in_edges(node) if incoming_edges else state.out_edges(node)
    return [describe_edge(e, incoming_edges, state) for e in edges]


def describe_incoming_edges(
    state: dace.SDFGState,
    node: dace_nodes.Node,
) -> list[EdgeConnectionSpec]:
    """Describes the incoming edges of `node`."""
    return describe_edges(state, node, True)


def describe_outgoing_edges(
    state: dace.SDFGState,
    node: dace_nodes.Node,
) -> list[EdgeConnectionSpec]:
    """Describes the out going edges of `node`."""
    return describe_edges(state, node, False)


def describe_all_edges(
    state: dace.SDFGState,
    node: dace_nodes.Node,
) -> list[EdgeConnectionSpec]:
    """Describes the all edges of `node`."""
    return describe_edges(state, node, False) + describe_edges(state, node, True)


def describes_incoming_edge(desc: EdgeConnectionSpec) -> bool:
    """Test if `desc` describes an incoming edge."""
    return desc.node is desc.edge.dst


def get_other_node(desc: EdgeConnectionSpec) -> dace_nodes.Node:
    """Extract the "other" node of the edge.

    Deprecated, use `desc.other_node` instead.
    """
    return desc.other_node


def get_other_subset(desc: EdgeConnectionSpec) -> dace_sbs.Subset:
    """Get the subset of the other side of the edge.

    Note, this is not the same as `desc.edge.data.other_subset`.
    Deprecated, use `desc.other_subset` instead.
    """
    return desc.other_subset


def are_intersecting(
    sbs1: dace_sbs.Subset,
    sbs2: dace_sbs.Subset,
) -> bool:
    """Returns `True` if `sbs1` and `sbs2` are intersecting for sure.

    The intersect test might look like as it has a binary outcome (intersecting
    or not intersecting), however, but there are three possible results:
    - Provable intersecting.
    - Provable not intersecting.
    - Unable to determine.

    This function will return `True` only if it can be shown that the two subsets
    intersects in all other cases it will return `True`.
    """
    return dace_sbs.intersects(sbs1, sbs2) is True


def never_intersecting(
    sbs1: dace_sbs.Subset,
    sbs2: dace_sbs.Subset,
) -> bool:
    """Returns `True` if `sbs1` and `sbs2` will never intersect.

    As described in `are_intersecting()` the intersection test can have
    three outcome. This function returns `True` if under no circumstance
    the two subsets will ever intersect. This means either it can be
    shown that the subsets will not intersect or it is impossible to
    determine the answer.
    """
    return not are_intersecting(sbs1, sbs2)


def maybe_intersecting(
    sbs1: dace_sbs.Subset,
    sbs2: dace_sbs.Subset,
) -> bool:
    """Returns `True` if `sbs1` and `sbs2` might be intersecting.

    As described in `are_intersecting()` there are three possible
    outcomes of the intersecting test. This function returns
    `True` if it is not possible to show that they are not intersecting.
    This means either they intersect or the question can not be
    answered.
    """
    return dace_sbs.intersects(sbs1, sbs2) is not False


def split_node(
    state: dace.SDFGState,
    sdfg: dace.SDFG,
    node_to_split: dace_nodes.AccessNode,
    split_description: Sequence[dace_sbs.Subset],
    allow_to_bypass_nodes: bool = False,
    already_reconfigured_nodes: Optional[set[tuple[dace_nodes.Node, str]]] = None,
) -> dict[dace_sbs.Subset, dace_nodes.AccessNode]:
    """The function will split `node_to_split` into several smaller AccessNodes.

    How the split is performed is described by `split_description`, which is a list
    of subsets that describes how to partition `node_to_split`.
    The function will then create new data descriptors for the for each of the fragment
    and create new AccessNodes. Then it will relocate all edges from `node_to_split`
    to the respective AccessNode of the fragment. The dataflow of the producer and
    consumer region will be reconfigured appropriate and strides are propagated.

    In some cases it might be possible to avoid to create an intermediate AccessNode.
    For example, if there is a Map that writes into `node_to_split` and a consumer
    to that node happens to be an AccessNode. In that case, the function will skip
    the write into the fragment and write directly into the consumer. However,
    for this `allow_to_bypass_nodes` must be `True`. Furthermore, note that the
    function will still create the data descriptors and the AccessNodes associated
    to the fragments.

    The function returns a `dict` that maps the subset describing each split to the
    AccessNode that was created for it.

    Note:
        - It is not required that`split_description` covers `node_to_split` completely.
            It is only required that a read or write is fully covered by a single
            fragment.
        - It is the responsibility of the caller to remove the isolated AccessNodes
            and remove the unused data descriptors.

    Todo:
        Make it possible to pass the AccessNodes and/or data descriptors from the
        outside.
    """
    if already_reconfigured_nodes is None:
        already_reconfigured_nodes = set()

    desc_to_split = node_to_split.desc(sdfg)
    assert desc_to_split.transient
    assert not gtx_transformations.utils.is_view(desc_to_split)
    assert isinstance(desc_to_split, dace_data.Array)

    input_descriptions = describe_incoming_edges(state, node_to_split)
    output_descriptions = describe_outgoing_edges(state, node_to_split)
    edge_descriptions = input_descriptions + output_descriptions

    # Ensure that no edge is on multiple new fragments and that every edge
    #  lands on a fragment.
    assert all(
        any(split.covers(desc.subset) for split in split_description) for desc in edge_descriptions
    ), "There is an edge that would be split."

    # Assign the edges of the node to a split.
    assignment = _compute_assignement_for_split(edge_descriptions, split_description)

    # Now we create the data descriptors and the AccessNodes that we need.
    new_data_descriptors = _generate_data_descriptors_for_split(
        state=state,
        sdfg=sdfg,
        node_to_split=node_to_split,
        split_description=split_description,
    )
    new_access_nodes: dict[dace_sbs.Subset, dace_nodes.AccessNode] = {
        split: state.add_access(dname, copy.copy(node_to_split.debuginfo))
        for split, dname in new_data_descriptors.items()
    }

    # This will also remove the `node_to_split` node but not the data descriptor
    #  it referred to.
    _perform_node_split(
        state=state,
        sdfg=sdfg,
        node_to_split=node_to_split,
        new_access_nodes=new_access_nodes,
        assignment=assignment,
        allow_to_bypass_nodes=allow_to_bypass_nodes,
        already_reconfigured_nodes=already_reconfigured_nodes,
    )

    return new_access_nodes


def split_edge(
    state: dace.SDFGState,
    sdfg: dace.SDFG,
    edge_to_split: dace_graph.MultiConnectorEdge,
    split_description: Sequence[dace_sbs.Subset],
) -> dict[Union[dace_sbs.Range, None], set[dace_graph.MultiConnectorEdge]]:
    """Tries to split `edge_to_split` into multiple edges.

    How the edge is split is described by `split_description`, which is a
    sequence of subsets. The function will decompose `edge_to_split` using
    `decompose_subset()` and then create a new edge for each resulting subset.

    The function returns a `dict` that maps a split subset to the set of
    edges that were created. Note that the `dict` also contains the
    key `None` which contains all the edges that are not associated to a
    split.

    It is important that this function is only able to split edges between
    AccessNodes.

    Args:
        state: The state in which we operate.
        sdfg: The SDFG on which we operate.
        edge_to_split: The edge that should be split.
        split_description: How the split should be executed.
    """

    # NOTE: The implementation is written from the point of view that
    #  `edge_to_split` is an outgoing edge. But this is an implementation
    #  detail that does not limit the applicability of this function.
    # TODO(phimuell): Implements some check that nothing is lost.

    assert isinstance(edge_to_split.src, dace_nodes.AccessNode)
    assert not isinstance(edge_to_split.src.desc(sdfg), dace_data.View)
    assert isinstance(edge_to_split.src.desc(sdfg), dace_data.Array)
    assert isinstance(edge_to_split.dst, dace_nodes.AccessNode)
    assert not isinstance(edge_to_split.dst.desc(sdfg), dace_data.View)
    assert isinstance(edge_to_split.dst.desc(sdfg), dace_data.Array)

    memlet_to_split: dace.Memlet = edge_to_split.data
    assert memlet_to_split.wcr is None
    assert memlet_to_split.subset is not None
    assert memlet_to_split.other_subset is not None
    assert all((r[2] == 1) == True for r in memlet_to_split.subset)  # noqa: E712 [true-false-comparison]  # SymPy comparison

    src: dace_nodes.AccessNode = edge_to_split.src
    consumer_subset: dace_sbs.Subset = memlet_to_split.src_subset

    # Fully split the consuming edge.
    # TODO(phimuell): Find a better way.
    fully_splitted_subsets: list[dace_sbs.Subset] = [consumer_subset]
    for split in split_description:
        new_fully_splitted_subsets: list[dace_sbs.Subset] = []
        for consumer in fully_splitted_subsets:
            split_res = decompose_subset(producer=split, consumer=consumer)
            if split_res:
                # It was possible to further split the consumer subset.
                new_fully_splitted_subsets.extend(split_res)
            else:
                # It was not possible to further split the consumer subset. Thus
                #  `consumer` is final.
                new_fully_splitted_subsets.append(consumer)
        fully_splitted_subsets = new_fully_splitted_subsets

    new_edges: dict[Union[dace_sbs.Range, None], dace_graph.MultiConnectorEdge] = {
        split: set() for split in split_description
    }
    new_edges[None] = set()

    consumer_dest_subset: dace_sbs.Subset = memlet_to_split.dst_subset
    consumer_dest_subset_start = consumer_dest_subset.min_element()
    consumer_subset_start = consumer_subset.min_element()
    for new_consumer_subset in fully_splitted_subsets:
        new_consumer_subset_start = new_consumer_subset.min_element()
        new_consumer_subset_size = new_consumer_subset.size()
        new_consumer_offset = [
            new_start - old_start
            for new_start, old_start in zip(new_consumer_subset_start, consumer_subset_start)
        ]

        new_consumer_dest_subset = dace_sbs.Range(
            [
                (old_dst_start + offset, old_dst_start + offset + size - 1, 1)
                for old_dst_start, size, offset in zip(
                    consumer_dest_subset_start, new_consumer_subset_size, new_consumer_offset
                )
            ]
        )
        new_edge = state.add_edge(
            src,
            edge_to_split.src_conn,
            edge_to_split.dst,
            edge_to_split.dst_conn,
            dace.Memlet(
                data=src.data,
                subset=new_consumer_subset,
                other_subset=new_consumer_dest_subset,
                dynamic=memlet_to_split.dynamic,
            ),
        )
        new_edges[
            next((split for split in split_description if split.covers(new_consumer_subset)), None)
        ].add(new_edge)

    state.remove_edge(edge_to_split)
    return new_edges


def decompose_subset(
    producer: dace_sbs.Subset,
    consumer: dace_sbs.Subset,
) -> Union[list[dace_sbs.Subset], None]:
    """
    Decompose `consumer` into pieces either covered by `producer` or have no intersection.

    If the decomposition exists the function returns a `list` containing the
    fragments, there are some special return values:
    - `None`: If the split is not applicable, for example there is no
        intersection in at least one dimensions.
    - The empty `list`: Indicates that `producer` fully covers `consumer`.

    Args:
        producer: The subset that can not be split.
        consumer: The subset that should be decomposed.
    """
    assert producer.dims() == consumer.dims()

    # Currently we require that we have to split only along one dimension.
    dimension_in_which_to_split: Optional[int] = None
    splitted_subsets_in_dim: list[tuple[Any, ...]] = []
    needs_further_spliting = False
    for dim in range(producer.dims()):
        prod_low = producer[dim][0]
        prod_high = producer[dim][1]
        consu_low = consumer[dim][0]
        consu_high = consumer[dim][1]

        # Check if the domains are the same.
        #  It seems that this must be a special case, at least it is handled that
        #  way in DaCe.
        equal_cond1 = (prod_low == consu_low) == True  # noqa: E712 [true-false-comparison]  # SymPy comparison
        equal_cond2 = (consu_high == prod_high) == True  # noqa: E712 [true-false-comparison]  # SymPy comparison
        if equal_cond1 and equal_cond2:
            continue

        # In this dimension the consumer consumes everything the producer
        #  generates. Therefore no splitting is needed.
        embedded_cond1 = (prod_low <= consu_low) == True  # noqa: E712 [true-false-comparison]  # SymPy comparison
        embedded_cond2 = (consu_high <= prod_high) == True  # noqa: E712 [true-false-comparison]  # SymPy comparison
        if embedded_cond1 and embedded_cond2:
            continue

        # Check if there is an intersection at all.
        #  I am pretty sure that there is no strange `-1` correction needed here.
        intersec_cond1 = consu_low <= prod_high
        intersec_cond2 = prod_low <= consu_high
        if intersec_cond1 == False or intersec_cond2 == False:  # noqa: E712 [true-false-comparison]  # SymPy comparison
            return None
        if not (intersec_cond1 == True and intersec_cond2 == True):  # noqa: E712 [true-false-comparison]  # SymPy comparison
            return None

        # `consumer` must be split in multiple dimensions, we already found one,
        #  stored in `dimension_in_which_to_split` but also found a new one, `dim`.
        #  We will now pretend that we did not see this. Instead we will do the
        #  split in the first dimension we found. We rely on the driver that it
        #  subsequently calls this function.
        if dimension_in_which_to_split is not None:
            needs_further_spliting = True
            continue
        else:
            dimension_in_which_to_split = dim

        # Determine the splitting case that we have.
        #  I am pretty sure about the `<` here.
        read_right = (prod_high < consu_high) == True  # noqa: E712 [true-false-comparison]  # SymPy comparison
        read_left = (consu_low < prod_low) == True  # noqa: E712 [true-false-comparison]  # SymPy comparison
        assert read_right or read_left

        # Now we determine the split mode. There are three cases.
        if read_right and read_left:
            # The consumer starts reading before the producer starts to write
            #  and also reads more than the producer writes to, so it is split
            #  into three parts.
            splitted_subsets_in_dim = [
                (consu_low, prod_low - 1, 1),
                (prod_low, prod_high, 1),
                (prod_high + 1, consu_high, 1),
            ]
        elif read_left:
            # The consumer starts reading before the producer starts writing.
            #  Thus there are two parts.
            splitted_subsets_in_dim = [(consu_low, prod_low - 1, 1), (prod_low, consu_high, 1)]
        elif read_right:
            # The consumer starts reading inside the range the producer writes to
            #  but reads more, so again two splits.
            splitted_subsets_in_dim = [
                (consu_low, prod_high, 1),
                (prod_high + 1, consu_high, 1),
            ]

    # If we are here this means that no split is applicable, because `consumer` is
    #  fully covered by `producer`. To indicate this we return the empty list.
    if dimension_in_which_to_split is None:
        assert not needs_further_spliting
        assert producer.covers(consumer)
        return []

    assert dimension_in_which_to_split is not None
    assert len(splitted_subsets_in_dim) > 0
    assert all(((e - s) >= 0) == True for s, e, _ in splitted_subsets_in_dim)  # noqa: E712 [true-false-comparison]  # SymPy comparison

    splitted_subsets: list[dace_sbs.Subset] = []
    for splitted_subset_in_dim in splitted_subsets_in_dim:
        splitted_subsets.append(
            dace_sbs.Range(
                [
                    (
                        splitted_subset_in_dim
                        if dim == dimension_in_which_to_split
                        else org_consumer_sbs
                    )
                    for dim, org_consumer_sbs in enumerate(consumer)
                ]
            )
        )

    # If we need further splitting then we must call the split subsets recursively.
    if needs_further_spliting:
        fully_splitted_subsets: list[dace_sbs.Subset] = []
        for consumer_fragment in splitted_subsets:
            next_decomposed_level = decompose_subset(producer=producer, consumer=consumer_fragment)
            if next_decomposed_level:
                fully_splitted_subsets.extend(next_decomposed_level)
            else:
                # There was nothing to split.
                fully_splitted_subsets.append(consumer_fragment)
        return fully_splitted_subsets
    else:
        return splitted_subsets


def subset_merger(
    subsets: Union[Sequence[EdgeConnectionSpec], Sequence[dace_sbs.Subset]],
) -> list[dace_sbs.Subset]:
    """Merges subsets together.

    The function tries the merge as many subsets together as possible.
    While, the function does not guarantees to find the largest possible
    subset, it guarantees that the result does not depend on the particular
    order of the passed subsets.

    Note:
        The function achieves its stability by first order the subsets in
        a particular order. For this it will serialize the subsets to strings
        and then sort them alphabetical. As this is the only way that also
        works in the presence of symbols.
    """
    assert len(subsets) > 0

    # Bring everything down to subsets. Note that it is important, that
    #  we do make a copy, because `_subset_merger_impl()` modifies its
    #  argument inplace.
    if isinstance(subsets[0], EdgeConnectionSpec):
        subsets = [copy.deepcopy(desc.subset) for desc in subsets]
    else:
        assert all(isinstance(sbs, dace_sbs.Subset) for sbs in subsets)
        subsets = [copy.deepcopy(sbs) for sbs in subsets]

    if len(subsets) == 1:
        return subsets

    return _subset_merger_impl(subsets)


def _subset_merger_impl(
    subsets: list[dace_sbs.Subset],
) -> list[dace_sbs.Subset]:
    """Implementation of the subset merger.

    The function will modify `subsets` inplace.
    """

    # The best we can do to ensure some kind of stability is sort them, however,
    #  in case of symbols we can not do that. Thus as the _only_ solution, we
    #  serialize them to string and order them accordingly.
    # NOTE: We can modify it inplace.
    subsets.sort(key=lambda sub: str(sub))

    performed_merge = True
    while performed_merge and (len(subsets) > 1):
        performed_merge = False

        # We could use `itertools.combinations()` here, but to guarantee deterministic
        #  processing, we will implement it yourselves.
        for idx1 in range(len(subsets)):
            for idx2 in range(idx1 + 1, len(subsets)):
                subset1, subset2 = subsets[idx1], subsets[idx2]
                merged_subset = _try_to_merge_subsets(subset1, subset2)
                if merged_subset is not None:
                    subsets.remove(subset1)
                    subsets.remove(subset2)
                    subsets.append(merged_subset)
                    performed_merge = True
                    break
            if performed_merge:
                break

    return subsets


def _try_to_merge_subsets(
    subset1: dace_sbs.Subset,
    subset2: dace_sbs.Subset,
) -> Union[None, dace_sbs.Subset]:
    """Tries to merge the subsets together, it it is impossible return `None`.

    Two subset can only be merged if they have the same bounds in all but one
    dimension. In that dimension the end index of one of the subset is the
    same as the start index of the other.
    """
    if subset1.dims() != subset2.dims():
        return None

    has_found_merge_dim = False
    merged_subset: list[dace_sym.SymbolicType] = []
    for dim in range(subset1.dims()):
        start1, end1, step1 = subset1[dim]
        start2, end2, step2 = subset2[dim]

        if (step1 != 1) == True or (step2 != 1) == True:  # noqa: E712 [true-false-comparison]  # SymPy comparison
            return None

        elif (start1 == start2) == True and (end1 == end2) == True:  # noqa: E712 [true-false-comparison]  # SymPy comparison
            merged_subset.append((start1, end1, 1))

        else:
            # We found a possible merge dimension.
            if has_found_merge_dim:
                # It is only possible to merge, actually extend, along in one dimension.
                return None

            if ((end1 + 1) == start2) == True:  # noqa: E712 [true-false-comparison]  # SymPy comparison
                merged_subset.append((start1, end2, 1))
            elif ((end2 + 1) == start1) == True:  # noqa: E712 [true-false-comparison]  # SymPy comparison
                merged_subset.append((start2, end1, 1))
            else:
                return None
            has_found_merge_dim = True

    return dace_sbs.Range(merged_subset)


def _perform_node_split(
    state: dace.SDFGState,
    sdfg: dace.SDFG,
    node_to_split: dace_nodes.AccessNode,
    new_access_nodes: dict[dace_sbs.Subset, dace_nodes.AccessNode],
    assignment: dict[dace_sbs.Subset, set[EdgeConnectionSpec]],
    allow_to_bypass_nodes: bool,
    already_reconfigured_nodes: set[tuple[dace_nodes.Node, str]],
) -> None:
    """Performs the actual split of `node_to_split` based on `assignment`.

    `assignment` describes how the edges of `node_to_split` should be distributed
    among the `new_access_nodes`. It maps a "split", i.e. a subset that describes
    one of the shards of the original AccessNode, to the edges that should be
    moved to another AccessNode, which is described through `new_access_nodes`.

    The function will also reconfigure the dataflow of the old edges. The argument
    `already_reconfigured_nodes` is used to keep track which dataflow has already
    been reconfigured.
    The function will then propagate the strides starting from the new functions.

    Furthermore, the function will remove the `node_to_split` AccessNode, but it
    will not remove the data.
    """
    assert all(sbs in assignment for sbs in new_access_nodes.keys())

    handled_edges: set[dace_graph.MultiConnectorEdge] = set()
    for split in new_access_nodes:
        edges_to_relocate = assignment[split]
        new_access_node = new_access_nodes[split]

        if len(edges_to_relocate) == 0:
            continue

        assert state.degree(new_access_node) == 0
        assert all(
            edge_to_relocate.edge not in handled_edges for edge_to_relocate in edges_to_relocate
        )

        if allow_to_bypass_nodes and _can_use_bypass_version_of_node_spliter(edges_to_relocate):
            _perform_node_split_with_bypass_impl(
                state=state,
                sdfg=sdfg,
                node_to_split=node_to_split,
                edges_to_relocate=edges_to_relocate,
                already_reconfigured_nodes=already_reconfigured_nodes,
            )
        else:
            for edge_to_relocate in edges_to_relocate:
                assert edge_to_relocate.edge not in handled_edges
                _perform_node_split_impl(
                    state=state,
                    sdfg=sdfg,
                    node_to_split=node_to_split,
                    new_access_node=new_access_node,
                    split_description=split,
                    edge_to_relocate=edge_to_relocate,
                    already_reconfigured_nodes=already_reconfigured_nodes,
                )
        handled_edges.update(edesc.edge for edesc in edges_to_relocate)

    # While we remove the AccessNode we do not remove the underlying data. We do this
    #  because it might be that there is another node that is still referring to
    #  it, which happens if we split across multiple states.
    assert state.degree(node_to_split) == 0
    state.remove_node(node_to_split)

    # Propagate the strides starting from the new access nodes.
    # NOTE: If the bypass version was used then the propagation has been done there
    #   already.
    for new_access_node in new_access_nodes.values():
        if state.degree(new_access_node) == 0:
            continue
        gtx_transformations.gt_propagate_strides_from_access_node(
            sdfg=sdfg,
            state=state,
            outer_node=new_access_node,
        )


def _perform_node_split_impl(
    state: dace.SDFGState,
    sdfg: dace.SDFG,
    node_to_split: dace_nodes.AccessNode,
    new_access_node: dace_nodes.AccessNode,
    split_description: dace_sbs.Subset,
    edge_to_relocate: EdgeConnectionSpec,
    already_reconfigured_nodes: set[tuple[dace_nodes.Node, str]],
) -> dace_graph.MultiConnectorEdge:
    """Performs the actual split.

    In essence this function will create a new edge, based on but `edge_to_relocate`,
    but where `node_to_split` is replaced with `new_access_node`. Furthermore, the
    subset of the edge is modified to reflect this change and the dataflow on the
    "other side of the edge" is also modified by means of `reconfigure_dataflow_after_rerouting()`.
    However, it is important that the function does not propagate strides.

    The old edge is removed and the new one is returned.
    """
    # TODO(phimuell): In case the other node is an AccessNode it is potentially
    #   possible to bypass `new_access_node`. This is an optimization that we
    #   should do as it also acts as kind of redundant array removal.

    # Because the edge no longer writes into `node_to_split` but into `new_access_node`,
    #  which is smaller, we have to adapt the size. However, because we assume that
    #  the split covers the subset the edge transfers, the offset is simply given
    #  by where the split starts. The minus is needed for the correct format in
    #  `reroute_edge()`.
    assert split_description.covers(edge_to_relocate.subset)
    subset_correction: Any = [-min_elem for min_elem in split_description.min_element()]

    is_producer_edge = describes_incoming_edge(edge_to_relocate)

    new_edge = gtx_transformations.utils.reroute_edge(
        is_producer_edge=is_producer_edge,
        current_edge=edge_to_relocate.edge,
        ss_offset=subset_correction,
        state=state,
        sdfg=sdfg,
        old_node=node_to_split,
        new_node=new_access_node,
    )

    # Depending on the situation we do not need a reconfiguration.
    if is_producer_edge:
        other_node: dace_nodes.Node = edge_to_relocate.edge.src
        other_node_conn = edge_to_relocate.edge.src_conn
    else:
        other_node = edge_to_relocate.edge.dst
        other_node_conn = edge_to_relocate.edge.dst_conn

    # Reconfigure the data flow on the other side.
    if (other_node, other_node_conn) not in already_reconfigured_nodes:
        already_reconfigured_nodes.add((other_node, other_node_conn))

        if isinstance(other_node, dace_nodes.NestedSDFG):
            # There is nothing special to do. In case of a nested SDFG, we will also have to do
            #  stride propagation, but we will postpone that.
            pass

        elif is_producer_edge and isinstance(other_node, dace_nodes.MapExit):
            pass

        elif (not is_producer_edge) and isinstance(other_node, dace_nodes.MapEntry):
            pass

        elif isinstance(other_node, dace_nodes.AccessNode):
            # Required by `reconfigure_dataflow_after_rerouting()`.
            if isinstance(other_node.desc(sdfg), dace_data.Scalar):
                subset_correction = None

        else:
            raise TypeError(f"Can not handle a producer of type '{type(other_node).__name__}'")

        gtx_transformations.utils.reconfigure_dataflow_after_rerouting(
            is_producer_edge=is_producer_edge,
            new_edge=new_edge,
            ss_offset=subset_correction,
            state=state,
            sdfg=sdfg,
            old_node=node_to_split,
            new_node=new_access_node,
        )

    state.remove_edge(edge_to_relocate.edge)
    return new_edge


def _can_use_bypass_version_of_node_spliter(
    edges_to_relocate: Iterable[EdgeConnectionSpec],
) -> bool:
    producer_edges: list[EdgeConnectionSpec] = [
        desc for desc in edges_to_relocate if describes_incoming_edge(desc)
    ]
    if len(producer_edges) != 1:
        return False
    if not isinstance(producer_edges[0].other_node, dace_nodes.AccessNode):
        return False
    return True


def _perform_node_split_with_bypass_impl(
    state: dace.SDFGState,
    sdfg: dace.SDFG,
    node_to_split: dace_nodes.AccessNode,
    edges_to_relocate: set[EdgeConnectionSpec],
    already_reconfigured_nodes: set[tuple[dace_nodes.Node, str]],
) -> list[dace_graph.MultiConnectorEdge]:
    """Performs the splitting but the edge might go directly to the consumer.

    # TODO: Remove the producer edge, run reconfiguration, split operation.
    # TODO ADDING PRODUCER TO THE SET OF PROCESSED NODES

    """
    producer_edge_desc = next(
        edesc for edesc in edges_to_relocate if describes_incoming_edge(edesc)
    )
    data_producer: dace_nodes.Node = producer_edge_desc.other_node
    producer_edge = producer_edge_desc.edge
    assert producer_edge.dst is node_to_split
    assert isinstance(data_producer, dace_nodes.AccessNode)

    old_producer_read = producer_edge.data.src_subset.min_element()
    old_producer_write = producer_edge.data.dst_subset.min_element()

    consumer_edges = [edesc for edesc in edges_to_relocate if edesc is not producer_edge_desc]
    new_consumer_edges: list[dace_graph.MultiConnectorEdge] = []
    for consumer_edge_desc in consumer_edges:
        consumer_edge = consumer_edge_desc.edge
        consumer_node: dace_nodes.Node = consumer_edge.dst
        consumer_conn = consumer_edge.dst_conn

        # Index from where the consumer should start reading from the producer
        #  directly. But since it has gone through the intermediate AccessNode,
        #  the indexes are different and we have to compute them. At the end it
        #  is some kind of projection starting at what was read from the
        #  intermediate to the indexes where it has been written to, to the
        #  intermediate and finally from where it is originally coming from.
        #  We only do this projection for the start index, the end index
        #  is computed by adding the length to the start index.
        old_consumer_read = consumer_edge.data.src_subset.min_element()
        consumer_read_size = consumer_edge.data.src_subset.size()
        consumer_direct_read: list[tuple[dace_sym.SymbolicType, dace_sym.SymbolicType, int]] = []
        for i in range(len(old_producer_read)):
            old_producer_read_start = old_producer_read[i]
            old_consumer_read_start = old_consumer_read[i]
            old_producer_write_start = old_producer_write[i]
            transfer_size = consumer_read_size[i]

            consumer_direct_read_start = dace.symbolic.pystr_to_symbolic(
                f"({old_producer_read_start}) + (({old_consumer_read_start}) - ({old_producer_write_start}))",
                simplify=True,
            )
            # The `-1` is because the end is considered inclusive in DaCe.
            consumer_direct_read_end = dace.symbolic.pystr_to_symbolic(
                f"({consumer_direct_read_start}) + ({transfer_size}) - 1", simplify=True
            )
            consumer_direct_read.append((consumer_direct_read_start, consumer_direct_read_end, 1))

        new_consumer_direct_read_subset = dace_sbs.Range(consumer_direct_read)

        # Create a new edge that reads from the producer directly and remove the
        #  old edge.
        new_consumer_edge = state.add_edge(
            producer_edge.src,
            producer_edge.src_conn,
            consumer_edge.dst,
            consumer_edge.dst_conn,
            dace.Memlet(
                data=data_producer.data,
                subset=new_consumer_direct_read_subset,
                other_subset=consumer_edge.data.dst_subset,
                dynamic=consumer_edge.data.dynamic or producer_edge.data.dynamic,
            ),
        )
        state.remove_edge(consumer_edge)
        new_consumer_edges.append(new_consumer_edge)

        # If needed reconfigure the consumers since it now involves the
        #  original producer.
        #  The stride propagation is done after all edges have been updated.
        if (consumer_node, consumer_conn) not in already_reconfigured_nodes:
            already_reconfigured_nodes.add((consumer_node, consumer_conn))

            # The subset correct we have to apply to the consumer depends on the
            #  type of the consumer.
            if isinstance(data_producer.desc(sdfg), dace_data.Scalar):
                # This is required by the reconfigure function.
                consumer_subset_correction = None

            elif isinstance(consumer_node, dace_nodes.AccessNode):
                # Here we only have to consider the offset between the reading
                #  from the intermediate and where we write into it. The minus
                #  is because we have to subtract this value.
                consumer_subset_correction = [
                    dace.symbolic.pystr_to_symbolic(
                        f"-(({old_consumer_read_start}) - ({old_producer_write_start}))",
                        simplify=True,
                    )
                    for old_consumer_read_start, old_producer_write_start in zip(
                        old_consumer_read,
                        old_producer_write,
                        strict=True,
                    )
                ]
            elif isinstance(consumer_node, dace_nodes.MapEntry):
                # Things are different here, because the map ranges (and the
                #  offsets in the inner Memlets handle everything, Thus, in
                #  this case we have to correct the case that they now read
                #  from the producer instead of the intermediate.
                consumer_subset_correction = [
                    dace.symbolic.pystr_to_symbolic(
                        f"-(({old_producer_write_start}) - ({old_producer_read_start}))",
                        simplify=True,
                    )
                    for old_producer_read_start, old_producer_write_start in zip(
                        old_producer_read,
                        old_producer_write,
                        strict=True,
                    )
                ]

            elif isinstance(consumer_node, dace_nodes.NestedSDFG):
                # Since a NestedSDFG can only read from an AccessNode there is
                #  nothing to do, except for the stride propagation, which is
                #  done later.
                continue

            else:
                raise TypeError(
                    f"Can not correct a consumer of type '{type(consumer_node).__name__}'"
                )

            gtx_transformations.utils.reconfigure_dataflow_after_rerouting(
                is_producer_edge=False,
                new_edge=new_consumer_edge,
                ss_offset=consumer_subset_correction,
                state=state,
                sdfg=sdfg,
                old_node=node_to_split,
                new_node=data_producer,
            )

    # We do not reconfigure the dataflow of the producer, because upstream of
    #  `data_producer` nothing has changed. In fact, we can not do it since
    #  there is also no edge. For that reason we do not add the producer to the
    #  `already_reconfigured_nodes` set, although one could do that.
    #  However, we have to propagate the strides, we have to do it here because on
    #  the outside it would only propagate from the new AccessNodes that we have
    #  bypassed.
    # TODO(phimuell): Find a way to avoid doing the propagation here, where the
    #   dataflow might be in some invalid state.
    state.remove_edge(producer_edge)
    processed_nsdfgs: set = set()
    for new_consumer_edge in new_consumer_edges:
        gtx_transformations.gt_map_strides_to_dst_nested_sdfg(
            outer_node=data_producer,
            edge=new_consumer_edge,
            sdfg=sdfg,
            state=state,
            processed_nsdfgs=processed_nsdfgs,
        )

    return new_consumer_edges


def _generate_data_descriptors_for_split(
    state: dace.SDFGState,
    sdfg: dace.SDFG,
    node_to_split: dace_nodes.AccessNode,
    split_description: Iterable[dace_sbs.Subset],
) -> dict[dace_sbs.Subset, str]:
    """Creates the data descriptor for every split.

    While the size is taken from `split_description` the other properties needed
    for the data descriptor are taken from `node_to_split`.
    """
    desc_to_split = node_to_split.desc(sdfg)
    new_data_descriptors: dict[dace_sbs.Subset, str] = {}

    for i, split_subset in enumerate(split_description):
        tmp_shape = split_subset.size()
        tmp_name, _ = sdfg.add_transient(
            name=f"{node_to_split.data}_split_{i}",
            shape=tmp_shape,
            dtype=desc_to_split.dtype,
            storage=desc_to_split.storage,
            find_new_name=True,
        )
        new_data_descriptors[split_subset] = tmp_name

    return new_data_descriptors


def _compute_assignement_for_split(
    edge_descriptions: Sequence[EdgeConnectionSpec],
    split_description: Sequence[dace_sbs.Subset],
) -> dict[dace_sbs.Subset, set[EdgeConnectionSpec]]:
    """For every subset, that defines a split find the set of edges that belongs into it.

    Note that it might happens that some splits have zero assigned edges.
    """
    assert all(split is not None for split in split_description)
    assignment: dict[dace_sbs.Subset, set[EdgeConnectionSpec]] = {
        split: set() for split in split_description
    }

    for edge_description in edge_descriptions:
        assert edge_description.subset is not None
        assigned_split = next(
            iter(split for split in split_description if split.covers(edge_description.subset)),
            None,
        )
        assert assigned_split is not None
        assignment[assigned_split].add(edge_description)

    return assignment



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Optional, Protocol

from gt4py.eve import utils as eve_utils
from gt4py.next import common
from gt4py.next.iterator import ir as itir
from gt4py.next.iterator.transforms import (
    concat_where,
    dead_code_elimination,
    fuse_as_fieldop,
    global_tmps,
    infer_domain,
    infer_domain_ops,
    inline_dynamic_shifts,
    inline_fundefs,
    inline_lifts,
    remove_broadcast,
)
from gt4py.next.iterator.transforms.collapse_list_get import CollapseListGet
from gt4py.next.iterator.transforms.collapse_tuple import CollapseTuple
from gt4py.next.iterator.transforms.constant_folding import ConstantFolding
from gt4py.next.iterator.transforms.cse import CommonSubexpressionElimination
from gt4py.next.iterator.transforms.fuse_maps import FuseMaps
from gt4py.next.iterator.transforms.inline_lambdas import InlineLambdas
from gt4py.next.iterator.transforms.inline_scalar import InlineScalar
from gt4py.next.iterator.transforms.merge_let import MergeLet
from gt4py.next.iterator.transforms.normalize_shifts import NormalizeShifts
from gt4py.next.iterator.transforms.unroll_reduce import UnrollReduce
from gt4py.next.iterator.type_system.inference import infer


class GTIRTransform(Protocol):
    def __call__(
        self, _: itir.Program, *, offset_provider: common.OffsetProvider
    ) -> itir.Program: ...


# TODO(tehrengruber): Revisit interface to configure temporary extraction. We currently forward
#  `extract_temporaries` and `temporary_extraction_heuristics` which is inconvenient.
def apply_common_transforms(
    ir: itir.Program,
    *,
    # TODO(havogt): should be replaced by `common.OffsetProviderType`, but global_tmps currently
    #  relies on runtime info or `symbolic_domain_sizes`.
    offset_provider: common.OffsetProvider | common.OffsetProviderType,
    extract_temporaries=False,
    unroll_reduce=False,
    common_subexpression_elimination=True,
    force_inline_lambda_args=False,
    unconditionally_collapse_tuples=False,
    #: A dictionary mapping axes names to their length. See :func:`infer_domain.infer_expr` for
    #: more details.
    symbolic_domain_sizes: Optional[dict[str, str]] = None,
) -> itir.Program:
    assert isinstance(ir, itir.Program)

    offset_provider_type = common.offset_provider_to_type(offset_provider)

    tmp_uids = eve_utils.UIDGenerator(prefix="__tmp")
    mergeasfop_uids = eve_utils.UIDGenerator()
    collapse_tuple_uids = eve_utils.UIDGenerator()

    ir = MergeLet().visit(ir)
    ir = inline_fundefs.InlineFundefs().visit(ir)

    ir = inline_fundefs.prune_unreferenced_fundefs(ir)
    ir = NormalizeShifts().visit(ir)

    # TODO(tehrengruber): Many iterator test contain lifts that need to be inlined, e.g.
    #  test_can_deref. We didn't notice previously as FieldOpFusion did this implicitly everywhere.
    ir = inline_lifts.InlineLifts().visit(ir)

    ir = dead_code_elimination.dead_code_elimination(
        ir, collapse_tuple_uids=collapse_tuple_uids, offset_provider_type=offset_provider_type
    )  # domain inference does not support dead-code
    ir = inline_dynamic_shifts.InlineDynamicShifts.apply(
        ir
    )  # domain inference does not support dynamic offsets yet
    ir = infer_domain_ops.InferDomainOps.apply(ir)
    ir = concat_where.canonicalize_domain_argument(ir)

    ir = concat_where.expand_tuple_args(ir, offset_provider_type=offset_provider_type)  # type: ignore[assignment]  # always an itir.Program
    ir = infer_domain.infer_program(
        ir,
        offset_provider=offset_provider,
        symbolic_domain_sizes=symbolic_domain_sizes,
    )
    ir = remove_broadcast.RemoveBroadcast.apply(ir)

    ir = concat_where.transform_to_as_fieldop(ir)

    for _ in range(10):
        inlined = ir

        inlined = InlineLambdas.apply(inlined, opcount_preserving=True)
        inlined = ConstantFolding.apply(inlined)  # type: ignore[assignment]  # always an itir.Program
        # This pass is required to be in the loop such that when an `if_` call with tuple arguments
        # is constant-folded the surrounding tuple_get calls can be removed.
        inlined = CollapseTuple.apply(
            inlined,
            enabled_transformations=~CollapseTuple.Transformation.PROPAGATE_TO_IF_ON_TUPLES,
            uids=collapse_tuple_uids,
            offset_provider_type=offset_provider_type,
        )  # type: ignore[assignment]  # always an itir.Program
        inlined = InlineScalar.apply(inlined, offset_provider_type=offset_provider_type)

        # This pass is required to run after CollapseTuple as otherwise we can not inline
        # expressions like `tuple_get(make_tuple(as_fieldop(stencil)(...)))` where stencil returns
        # a list. Such expressions must be inlined however because no backend supports such
        # field operators right now.
        inlined = fuse_as_fieldop.FuseAsFieldOp.apply(
            inlined, uids=mergeasfop_uids, offset_provider_type=offset_provider_type
        )

        if inlined == ir:
            break
        ir = inlined
    else:
        raise RuntimeError("Inlining 'lift' and 'lambdas' did not converge.")

    # breaks in test_zero_dim_tuple_arg as trivial tuple_get is not inlined
    if common_subexpression_elimination:
        ir = CommonSubexpressionElimination.apply(ir, offset_provider_type=offset_provider_type)
        ir = MergeLet().visit(ir)
        ir = InlineLambdas.apply(ir, opcount_preserving=True)

    if extract_temporaries:
        ir = infer(ir, inplace=True, offset_provider_type=offset_provider_type)
        ir = global_tmps.create_global_tmps(
            ir,
            offset_provider=offset_provider,
            symbolic_domain_sizes=symbolic_domain_sizes,
            uids=tmp_uids,
        )

    # Since `CollapseTuple` relies on the type inference which does not support returning tuples
    # larger than the number of closure outputs as given by the unconditional collapse, we can
    # only run the unconditional version here instead of in the loop above.
    if unconditionally_collapse_tuples:
        ir = CollapseTuple.apply(
            ir,
            ignore_tuple_size=True,
            uids=collapse_tuple_uids,
            enabled_transformations=~CollapseTuple.Transformation.PROPAGATE_TO_IF_ON_TUPLES,
            offset_provider_type=offset_provider_type,
        )  # type: ignore[assignment]  # always an itir.Program

    ir = NormalizeShifts().visit(ir)

    ir = FuseMaps().visit(ir)
    ir = CollapseListGet().visit(ir)

    if unroll_reduce:
        for _ in range(10):
            unrolled = UnrollReduce.apply(ir, offset_provider_type=offset_provider_type)
            unrolled = CollapseListGet().visit(unrolled)
            unrolled = NormalizeShifts().visit(unrolled)
            # this is required as nested neighbor reductions can contain lifts, e.g.,
            # `neighbors(V2Eₒ, ↑f(...))`
            unrolled = inline_lifts.InlineLifts().visit(unrolled)
            unrolled = NormalizeShifts().visit(unrolled)
            if unrolled == ir:
                break
            ir = unrolled  # type: ignore[assignment] # still a `itir.Program`
        else:
            raise RuntimeError("Reduction unrolling failed.")

    ir = InlineLambdas.apply(
        ir, opcount_preserving=True, force_inline_lambda_args=force_inline_lambda_args
    )

    assert isinstance(ir, itir.Program)
    return ir


def apply_fieldview_transforms(
    ir: itir.Program, *, offset_provider: common.OffsetProvider
) -> itir.Program:
    offset_provider_type = common.offset_provider_to_type(offset_provider)

    ir = inline_fundefs.InlineFundefs().visit(ir)
    ir = inline_fundefs.prune_unreferenced_fundefs(ir)
    ir = dead_code_elimination.dead_code_elimination(ir, offset_provider_type=offset_provider_type)
    ir = inline_dynamic_shifts.InlineDynamicShifts.apply(
        ir
    )  # domain inference does not support dynamic offsets yet

    ir = infer_domain_ops.InferDomainOps.apply(ir)
    ir = concat_where.canonicalize_domain_argument(ir)
    ir = ConstantFolding.apply(ir)  # type: ignore[assignment]  # always an itir.Program

    ir = infer_domain.infer_program(ir, offset_provider=offset_provider)
    ir = remove_broadcast.RemoveBroadcast.apply(ir)
    return ir



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause


from gt4py.next.type_system import type_specifications as ts


def pytype_to_cpptype(t: ts.ScalarType | str) -> str:
    if isinstance(t, ts.ScalarType):
        t = t.kind.name.lower()
    try:
        return {
            "float32": "float",
            "float64": "double",
            "int8": "std::int8_t",
            "uint8": "std::uint8_t",
            "int16": "std::int16_t",
            "uint16": "std::uint16_t",
            "int32": "std::int32_t",
            "uint32": "std::uint32_t",
            "int64": "std::int64_t",
            "uint64": "std::uint64_t",
            "bool": "bool",
            "string": "string",
        }[t]
    except KeyError:
        raise TypeError(f"Unsupported type '{t}'.") from None



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from gt4py.next.iterator.builtins import reduce


def sum_(fun=None):
    if fun is None:
        return reduce(lambda a, b: a + b, 0.0)
    else:
        return reduce(lambda first, a, b: first + fun(a, b), 0.0)  # TODO tracing for *args


def dot(a, b):
    return reduce(lambda acc, a_n, b_n: acc + a_n * b_n, 0.0)(a, b)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import operator
from dataclasses import dataclass
from functools import reduce
from typing import Any, Final

from dace import Memlet, nodes, subsets

from gt4py import eve
from gt4py.cartesian.gtc import common, oir
from gt4py.cartesian.gtc.dace import treeir as tir, utils


# Tasklet in/out connector prefixes
TASKLET_IN: Final[str] = "gtIN_"
TASKLET_OUT: Final[str] = "gtOUT_"


@dataclass
class Context:
    code: list[str]
    """Tasklet code, line by line."""

    targets: set[str]
    """Names of fields / scalars that we've already written to. Used for read-after-write analysis."""

    inputs: dict[str, Memlet]
    """Mapping connector names to memlets flowing into the Tasklet."""

    outputs: dict[str, Memlet]
    """Mapping connector names to memlets flowing out of the Tasklet."""

    tree: tir.TreeRoot
    """Schedule tree in which this Tasklet will be inserted."""


class OIRToTasklet(eve.NodeVisitor):
    """
    Translate the numerical code from OIR to DaCe Tasklets.

    This visitor should neither attempt transformations nor do any control flow
    work. Control flow is the responsibility of OIRToTreeIR.
    """

    def visit_CodeBlock(
        self, node: oir.CodeBlock, root: tir.TreeRoot
    ) -> tuple[nodes.Tasklet, dict[str, Memlet], dict[str, Memlet]]:
        """Entry point to gather all code, inputs and outputs."""
        ctx = Context(code=[], targets=set(), inputs={}, outputs={}, tree=root)

        self.visit(node.body, ctx=ctx)

        tasklet = nodes.Tasklet(
            label=node.label,
            code="\n".join(ctx.code),
            inputs=ctx.inputs.keys(),
            outputs=ctx.outputs.keys(),
        )

        return (tasklet, ctx.inputs, ctx.outputs)

    def visit_ScalarAccess(self, node: oir.ScalarAccess, ctx: Context, is_target: bool) -> str:
        target = is_target or node.name in ctx.targets
        tasklet_name = _tasklet_name(node, target)

        if (
            node.name in ctx.targets  # (read or write) after write
            or tasklet_name in ctx.inputs  # read after read
        ):
            return tasklet_name

        memlet = Memlet(data=node.name, subset=subsets.Range([(0, 0, 1)]))
        if is_target:
            # Note: it doesn't matter if we use is_target or target here because if they
            # were different, we had a read-after-write situation, which was already
            # handled above.
            ctx.targets.add(node.name)
            ctx.outputs[tasklet_name] = memlet
        else:
            ctx.inputs[tasklet_name] = memlet

        return tasklet_name

    def visit_FieldAccess(self, node: oir.FieldAccess, ctx: Context, is_target: bool) -> str:
        # Derive tasklet name of this access
        postfix = _field_offset_postfix(node)
        key = f"{node.name}_{postfix}"
        target = is_target or key in ctx.targets
        tasklet_name = _tasklet_name(node, target, postfix)

        # Gather all parts of the variable name in this list
        name_parts = [tasklet_name]

        # Variable K offset subscript
        if isinstance(node.offset, oir.VariableKOffset):
            symbol = tir.Axis.K.iteration_dace_symbol()
            shift = ctx.tree.shift[node.name][tir.Axis.K]
            offset = self.visit(node.offset.k, ctx=ctx, is_target=False)
            name_parts.append(f"[({symbol}) + ({shift}) + ({offset})]")

        # Data dimension subscript
        data_indices: list[str] = []
        for index in node.data_index:
            data_indices.append(self.visit(index, ctx=ctx, is_target=False))

        if data_indices:
            name_parts.append(f"[{', '.join(data_indices)}]")

        # In case this is the second access (inside the same tasklet), we can just return the
        # name and don't have to build a Memlet anymore.
        if (
            key in ctx.targets  # (read or write) after write
            or tasklet_name in ctx.inputs  # read after read
        ):
            return "".join(filter(None, name_parts))

        # Build Memlet and add it to inputs/outputs
        data_domains: list[int] = (
            ctx.tree.containers[node.name].shape[-len(node.data_index) :] if node.data_index else []
        )
        memlet = Memlet(
            data=node.name,
            subset=_memlet_subset(node, data_domains, ctx),
            volume=reduce(operator.mul, data_domains, 1),  # correct volume for VariableK offsets
        )
        if is_target:
            # Note: it doesn't matter if we use is_target or target here because if they
            # were different, we had a read-after-write situation, which was already
            # handled above.
            ctx.targets.add(key)
            ctx.outputs[tasklet_name] = memlet
        else:
            ctx.inputs[tasklet_name] = memlet

        return "".join(filter(None, name_parts))

    def visit_AssignStmt(self, node: oir.AssignStmt, ctx: Context) -> None:
        # Order matters: always evaluate the right side of an assignment first
        right = self.visit(node.right, ctx=ctx, is_target=False)
        left = self.visit(node.left, ctx=ctx, is_target=True)

        ctx.code.append(f"{left} = {right}")

    def visit_TernaryOp(self, node: oir.TernaryOp, **kwargs: Any) -> str:
        condition = self.visit(node.cond, **kwargs)
        if_code = self.visit(node.true_expr, **kwargs)
        else_code = self.visit(node.false_expr, **kwargs)

        return f"({if_code} if {condition} else {else_code})"

    def visit_BinaryOp(self, node: oir.BinaryOp, **kwargs: Any) -> str:
        left = self.visit(node.left, **kwargs)
        right = self.visit(node.right, **kwargs)

        return f"({left} {node.op.value} {right})"

    def visit_UnaryOp(self, node: oir.UnaryOp, **kwargs: Any) -> str:
        expr = self.visit(node.expr, **kwargs)

        return f"{node.op.value}({expr})"

    def visit_Cast(self, node: oir.Cast, **kwargs: Any) -> str:
        dtype = utils.data_type_to_dace_typeclass(node.dtype)
        expression = self.visit(node.expr, **kwargs)

        return f"{dtype}({expression})"

    def visit_Literal(self, node: oir.Literal, **kwargs: Any) -> str:
        if isinstance(node.value, common.BuiltInLiteral):
            return self.visit(node.value, **kwargs)

        # Resolve int and float literals to the correct precision
        dtype = utils.data_type_to_dace_typeclass(node.dtype)
        return f"{dtype}({node.value})"

    def visit_BuiltInLiteral(self, node: common.BuiltInLiteral, **_kwargs: Any) -> str:
        if node == common.BuiltInLiteral.TRUE:
            return "True"

        if node == common.BuiltInLiteral.FALSE:
            return "False"

        raise NotImplementedError(f"BuiltInLiteral '{node}' not (yet) implemented.")

    def visit_NativeFunction(self, node: common.NativeFunction, **_kwargs: Any) -> str:
        native_functions = {
            common.NativeFunction.ABS: "abs",
            common.NativeFunction.MIN: "min",
            common.NativeFunction.MAX: "max",
            common.NativeFunction.MOD: "fmod",
            common.NativeFunction.SIN: "dace.math.sin",
            common.NativeFunction.COS: "dace.math.cos",
            common.NativeFunction.TAN: "dace.math.tan",
            common.NativeFunction.ARCSIN: "asin",
            common.NativeFunction.ARCCOS: "acos",
            common.NativeFunction.ARCTAN: "atan",
            common.NativeFunction.SINH: "dace.math.sinh",
            common.NativeFunction.COSH: "dace.math.cosh",
            common.NativeFunction.TANH: "dace.math.tanh",
            common.NativeFunction.ARCSINH: "asinh",
            common.NativeFunction.ARCCOSH: "acosh",
            common.NativeFunction.ARCTANH: "atanh",
            common.NativeFunction.SQRT: "dace.math.sqrt",
            common.NativeFunction.POW: "dace.math.pow",
            common.NativeFunction.EXP: "dace.math.exp",
            common.NativeFunction.LOG: "dace.math.log",
            common.NativeFunction.LOG10: "log10",
            common.NativeFunction.GAMMA: "tgamma",
            common.NativeFunction.CBRT: "cbrt",
            common.NativeFunction.ISFINITE: "isfinite",
            common.NativeFunction.ISINF: "isinf",
            common.NativeFunction.ISNAN: "isnan",
            common.NativeFunction.FLOOR: "dace.math.ifloor",
            common.NativeFunction.CEIL: "ceil",
            common.NativeFunction.TRUNC: "trunc",
            common.NativeFunction.INT32: "dace.int32",
            common.NativeFunction.INT64: "dace.int64",
            common.NativeFunction.FLOAT32: "dace.float32",
            common.NativeFunction.FLOAT64: "dace.float64",
            common.NativeFunction.ERF: "erf",
            common.NativeFunction.ERFC: "erfc",
        }
        if node not in native_functions:
            raise NotImplementedError(f"NativeFunction '{node}' not (yet) implemented.")

        return native_functions[node]

    def visit_NativeFuncCall(self, node: oir.NativeFuncCall, **kwargs: Any) -> str:
        function_name = self.visit(node.func, **kwargs)
        arguments = ",".join([self.visit(a, **kwargs) for a in node.args])

        return f"{function_name}({arguments})"

    # Not (yet) supported section
    def visit_CacheDesc(self, node: oir.CacheDesc, **kwargs: Any) -> None:
        raise NotImplementedError("To be implemented: Caches")

    def visit_IJCache(self, node: oir.IJCache, **kwargs: Any) -> None:
        raise NotImplementedError("To be implemented: Caches")

    def visit_KCache(self, node: oir.KCache, **kwargs: Any) -> None:
        raise NotImplementedError("To be implemented: Caches")

    # Should _not_ be called
    def visit_CartesianOffset(self, node: common.CartesianOffset, **kwargs: Any) -> None:
        raise RuntimeError("Cartesian Offset should be dealt in Access IRs.")

    def visit_VariableKOffset(self, node: oir.VariableKOffset, **kwargs: Any) -> None:
        raise RuntimeError("Variable K Offset should be dealt in Access IRs.")

    def visit_MaskStmt(self, node: oir.MaskStmt, **kwargs: Any) -> None:
        raise RuntimeError("visit_MaskStmt should not be called")

    def visit_While(self, node: oir.While, **kwargs: Any) -> None:
        raise RuntimeError("visit_While should not be called")

    def visit_HorizontalRestriction(self, node: oir.HorizontalRestriction, **kwargs: Any) -> None:
        raise RuntimeError("visit_HorizontalRestriction: should be dealt in TreeIR")

    def visit_LocalScalar(self, node: oir.LocalScalar, **kwargs: Any) -> None:
        raise RuntimeError("visit_LocalScalar should not be called")

    def visit_Temporary(self, node: oir.Temporary, **kwargs: Any) -> None:
        raise RuntimeError("visit_LocalScalar should not be called")

    def visit_Stencil(self, node: oir.Stencil, **kwargs: Any) -> None:
        raise RuntimeError("visit_Stencil should not be called")

    def visit_Decl(self, node: oir.Decl, **kwargs: Any) -> None:
        raise RuntimeError("visit_Decl should not be called")

    def visit_FieldDecl(self, node: oir.FieldDecl, **kwargs: Any) -> None:
        raise RuntimeError("visit_FieldDecl should not be called")

    def visit_ScalarDecl(self, node: oir.ScalarDecl, **kwargs: Any) -> None:
        raise RuntimeError("visit_ScalarDecl should not be called")

    def visit_Interval(self, node: oir.Interval, **kwargs: Any) -> None:
        raise RuntimeError("visit_Interval should not be called")

    def visit_UnboundedInterval(self, node: oir.UnboundedInterval, **kwargs: Any) -> None:
        raise RuntimeError("visit_UnboundedInterval should not be called")

    def visit_HorizontalExecution(self, node: oir.HorizontalExecution, **kwargs: Any) -> None:
        raise RuntimeError("visit_HorizontalExecution should not be called")

    def visit_VerticalLoop(self, node: oir.VerticalLoop, **kwargs: Any) -> None:
        raise RuntimeError("visit_VerticalLoop should not be called")

    def visit_VerticalLoopSection(self, node: oir.VerticalLoopSection, **kwargs: Any) -> None:
        raise RuntimeError("visit_VerticalLoopSection should not be called")


def _tasklet_name(
    node: oir.FieldAccess | oir.ScalarAccess, is_target: bool, postfix: str = ""
) -> str:
    name_prefix = TASKLET_OUT if is_target else TASKLET_IN
    return "_".join(filter(None, [name_prefix, node.name, postfix]))


def _field_offset_postfix(node: oir.FieldAccess) -> str:
    if isinstance(node.offset, oir.VariableKOffset):
        return "var_k"

    offset_indicators = [
        f"{k}{'p' if v > 0 else 'm'}{abs(v)}" for k, v in node.offset.to_dict().items() if v != 0
    ]
    return "_".join(offset_indicators)


def _memlet_subset(node: oir.FieldAccess, data_domains: list[int], ctx: Context) -> subsets.Subset:
    if isinstance(node.offset, common.CartesianOffset):
        return _memlet_subset_cartesian(node, data_domains, ctx)

    if isinstance(node.offset, oir.VariableKOffset):
        return _memlet_subset_variable_offset(node, data_domains, ctx)

    raise NotImplementedError(f"_memlet_subset(): unknown offset type {type(node.offset)}")


def _memlet_subset_cartesian(
    node: oir.FieldAccess, data_domains: list[int], ctx: Context
) -> subsets.Subset:
    """
    Generates the memlet subset for a field access with a cartesian offset.

    Note that we pass data dimensions as a full array into the Tasklet. For cases with data dimensions
    we thus need a Range subset. We could use the more narrow Indices subset for cases without data
    dimensions. For the sake of simplicity, we choose to always return a Range.
    """
    offset_dict = node.offset.to_dict()
    dimensions = ctx.tree.dimensions[node.name]
    shift = ctx.tree.shift[node.name]

    ranges: list[tuple[str | int, str | int, int]] = []
    # Handle cartesian indices
    for index, axis in enumerate(tir.Axis.dims_3d()):
        if dimensions[index]:
            i = f"({axis.iteration_dace_symbol()}) + ({shift[axis]}) + ({offset_dict[axis.lower()]})"
            ranges.append((i, i, 1))

    # Append data dimensions
    for domain_size in data_domains:
        ranges.append((0, domain_size - 1, 1))  # ranges are inclusive

    return subsets.Range(ranges)


def _memlet_subset_variable_offset(
    node: oir.FieldAccess, data_domains: list[int], ctx: Context
) -> subsets.Subset:
    """
    Generates the memlet subset for a field access with a variable K offset.

    While we know that we are reading at one specific i/j/k access, the K-access point is only
    determined at runtime. We thus pass the K-axis as array into the Tasklet.
    """
    # Handle cartesian indices
    shift = ctx.tree.shift[node.name]
    offset_dict = node.offset.to_dict()
    i = f"({tir.Axis.I.iteration_symbol()}) + ({shift[tir.Axis.I]}) + ({offset_dict[tir.Axis.I.lower()]})"
    j = f"({tir.Axis.J.iteration_symbol()}) + ({shift[tir.Axis.J]}) + ({offset_dict[tir.Axis.J.lower()]})"
    K = f"({tir.Axis.K.domain_symbol()}) + ({shift[tir.Axis.K]}) - 1"  # ranges are inclusive
    ranges: list[tuple[str | int, str | int, int]] = [(i, i, 1), (j, j, 1), (0, K, 1)]

    # Append data dimensions
    for domain_size in data_domains:
        ranges.append((0, domain_size - 1, 1))  # ranges are inclusive

    return subsets.Range(ranges)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import abc
import copy
import hashlib
import os
import pathlib
import time
import warnings
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Protocol

from typing_extensions import deprecated

from gt4py import storage as gt_storage
from gt4py.cartesian import definitions as gt_definitions, utils as gt_utils
from gt4py.cartesian.backend import pyext_builder
from gt4py.cartesian.backend.module_generator import (
    BaseModuleGenerator,
    ModuleData,
    make_args_data_from_gtir,
)


if TYPE_CHECKING:
    from gt4py.cartesian.stencil_builder import StencilBuilder
    from gt4py.cartesian.stencil_object import StencilObject


class Backend(abc.ABC):
    #: Backend name
    name: ClassVar[str]

    #: Backend-specific options:
    #:
    #:  + info:
    #:    - versioning: is versioning on?
    #:    - description [optional]
    #:    - type
    options: ClassVar[dict[str, Any]]

    #: Backend-specific storage parametrization:
    #:
    #:  - "alignment": in bytes
    #:  - "device": "cpu" | "gpu"
    #:  - "layout_map": callback converting a mask to a layout
    #:  - "is_optimal_layout": callback checking if a storage has compatible layout
    storage_info: ClassVar[gt_storage.layout.LayoutInfo]

    #: Language support:
    #:
    #:  - "computation": name of the language in which the computation is implemented.
    #:  - "bindings": names of supported language bindings / wrappers.
    #:    If a high-level language is compatible out-of-the-box, it should not be listed.
    #:
    #:  Languages should be spelled using the official spelling
    #:  but lower case ("python", "fortran", "rust").
    languages: ClassVar[dict[str, Any] | None] = None

    # __impl_opts:
    #   "disable-code-generation": bool
    #   "disable-cache-validation": bool

    builder: StencilBuilder

    def __init__(self, builder: StencilBuilder) -> None:
        self.builder = builder

    @classmethod
    def filter_options_for_id(
        cls, options: gt_definitions.BuildOptions
    ) -> gt_definitions.BuildOptions:
        """Filter a copy for options that should not be part of the stencil ID."""
        filtered_options = copy.deepcopy(options)
        id_names = set(name for name, info in cls.options.items() if info["versioning"])
        non_id_names = set(filtered_options.backend_opts.keys()) - id_names
        for name in non_id_names:
            del filtered_options.backend_opts[name]

        return filtered_options

    @abc.abstractmethod
    def load(self) -> type[StencilObject] | None:
        """
        Load the stencil class from the generated python module.

        type:
            The generated stencil class after loading through python's import API


        This assumes that :py:meth:`Backend.generate` has been called already on the same stencil.
        The python module therefore already exists and any python extensions have been compiled.
        In case the stencil changed since the last generate call, it will not be rebuilt.
        """
        pass

    @abc.abstractmethod
    def generate(self) -> type[StencilObject]:
        """
        Generate the stencil class from GTScript's internal representation.

        Returns
        -------
        type:
            The generated stencil class after loading through python's import API

        In case the stencil class for the given ID is found in cache, this method can avoid
        rebuilding it. Rebuilding can however be triggered through the :code:`options` parameter
        if supported by the backend.
        """
        pass

    @property
    def extra_cache_info(self) -> dict[str, Any]:
        """Provide additional data to be stored in cache info file (subclass hook)."""
        return {}

    @property
    def extra_cache_validation_keys(self) -> list[str]:
        """List keys from extra_cache_info to be validated during consistency check."""
        return []


REGISTRY = gt_utils.Registry[type[Backend]]()


def from_name(name: str) -> type[Backend]:
    """Return a backend by name."""
    backend_cls = REGISTRY.get(name, None)
    if backend_cls is None:
        raise NotImplementedError(
            f"Backend '{name}' is not implemented. Options are: {REGISTRY.names}."
        )
    return backend_cls


def register(backend_cls: type[Backend]) -> type[Backend]:
    """Register a backend."""
    assert issubclass(backend_cls, Backend) and backend_cls.name is not None

    if isinstance(backend_cls.name, str):
        gt_storage.register(backend_cls.name, backend_cls.storage_info)
        return REGISTRY.register(backend_cls.name, backend_cls)

    raise ValueError(
        f"Invalid 'name' attribute ('{backend_cls.name}') in backend class '{backend_cls}'."
    )


class BaseBackend(Backend):
    MODULE_GENERATOR_CLASS: ClassVar[type[BaseModuleGenerator]]

    def load(self) -> type[StencilObject] | None:
        build_info = self.builder.options.build_info
        if build_info is not None:
            start_time = time.perf_counter()

        stencil_class = None
        if self.builder.stencil_id is not None:
            self.check_options(self.builder.options)
            validate_hash = not self.builder.options._impl_opts.get(
                "disable-cache-validation", False
            )
            if self.builder.caching.is_cache_info_available_and_consistent(
                validate_hash=validate_hash
            ):
                stencil_class = self._load()

        if build_info is not None:
            build_info["load_time"] = time.perf_counter() - start_time

        return stencil_class

    def generate(self) -> type[StencilObject]:
        self.check_options(self.builder.options)
        return self.make_module()

    def _load(self) -> type[StencilObject]:
        stencil_class_name = self.builder.class_name
        file_name = str(self.builder.module_path)
        stencil_module = gt_utils.make_module_from_file(stencil_class_name, file_name)
        stencil_class: type[StencilObject] = getattr(stencil_module, stencil_class_name)
        stencil_class.__module__ = self.builder.module_qualname
        stencil_class._gt_id_ = self.builder.stencil_id.version
        stencil_class.definition_func = staticmethod(self.builder.definition)

        return stencil_class

    def check_options(self, options: gt_definitions.BuildOptions) -> None:
        assert self.options is not None
        unknown_options = set(options.backend_opts.keys()) - set(self.options.keys())
        if unknown_options:
            warnings.warn(
                f"Unknown options '{unknown_options}' for backend '{self.name}'",
                RuntimeWarning,
                stacklevel=2,
            )

    def make_module(self) -> type[StencilObject]:
        build_info = self.builder.options.build_info
        if build_info is not None:
            start_time = time.perf_counter()

        file_path = self.builder.module_path
        module_source = self.make_module_source()

        if not self.builder.options._impl_opts.get("disable-code-generation", False):
            file_path.parent.mkdir(parents=True, exist_ok=True)
            file_path.write_text(module_source)
            self.builder.caching.update_cache_info()

        module = self._load()

        if build_info is not None:
            build_info["module_time"] = time.perf_counter() - start_time

        return module

    def make_module_source(self, *, args_data: ModuleData | None = None) -> str:
        """Generate the module source code with or without stencil id."""
        args_data = args_data or make_args_data_from_gtir(self.builder.gtir_pipeline)
        return self.MODULE_GENERATOR_CLASS(self.builder)(args_data)


class MakeModuleSourceCallable(Protocol):
    def __call__(self, *, args_data: ModuleData | None = None) -> str: ...


class BasePyExtBackend(BaseBackend):
    @property
    def pyext_module_name(self) -> str:
        return self.builder.module_name + "_pyext"

    @property
    def pyext_module_path(self) -> str:
        return self.builder.module_qualname + "_pyext"

    @property
    def pyext_class_name(self) -> str:
        return self.builder.class_name + "_pyext"

    @property
    def pyext_build_dir_path(self) -> pathlib.Path:
        return self.builder.pkg_path.joinpath(self.pyext_module_name + "_BUILD")

    @property
    def extra_cache_info(self) -> dict[str, Any]:
        pyext_file_path = self.builder.backend_data.get("pyext_file_path", None)
        pyext_md5 = ""
        if pyext_file_path:
            pyext_md5 = hashlib.md5(pathlib.Path(pyext_file_path).read_bytes()).hexdigest()
        return {
            **super().extra_cache_info,
            "pyext_file_path": pyext_file_path,
            "pyext_md5": pyext_md5,
        }

    @property
    def extra_cache_validation_keys(self) -> list[str]:
        keys = super().extra_cache_validation_keys
        if self.extra_cache_info["pyext_md5"]:
            keys.append("pyext_md5")
        return keys

    @abc.abstractmethod
    def generate(self) -> type[StencilObject]:
        pass

    def build_extension_module(
        self,
        pyext_sources: dict[str, Any],
        pyext_build_opts: dict[str, str],
        *,
        uses_cuda: bool = False,
    ) -> None:
        # Build extension module
        pyext_build_path = pathlib.Path(
            os.path.relpath(self.pyext_build_dir_path, pathlib.Path.cwd())
        )
        pyext_build_path.mkdir(parents=True, exist_ok=True)
        sources = []
        for key, source in pyext_sources.items():
            src_file_path = pyext_build_path / key
            src_ext = src_file_path.suffix
            if src_ext not in [".h", ".hpp"]:
                sources.append(str(src_file_path))

            if source is not gt_utils.NOTHING:
                src_file_path.write_text(source)

        pyext_target_file_path = self.builder.pkg_path
        qualified_pyext_name = self.pyext_module_path

        pyext_build_args: dict[str, Any] = dict(
            name=qualified_pyext_name,
            sources=sources,
            build_path=str(pyext_build_path),
            target_path=str(pyext_target_file_path),
            **pyext_build_opts,
        )

        if uses_cuda:
            module_name, file_path = pyext_builder.build_pybind_cuda_ext(**pyext_build_args)
        else:
            module_name, file_path = pyext_builder.build_pybind_ext(**pyext_build_args)

        assert module_name == qualified_pyext_name

        self.builder.with_backend_data(
            {"pyext_module_name": module_name, "pyext_file_path": file_path}
        )


def disabled(message: str, *, enabled_env_var: str) -> Callable[[type[Backend]], type[Backend]]:
    # We push for hard deprecation here by raising by default and warning if enabling has been forced.
    enabled = bool(int(os.environ.get(enabled_env_var, "0")))
    if enabled:
        return deprecated(message)

    def _decorator(cls: type[Backend]) -> type[Backend]:
        def _no_generate(obj) -> type[StencilObject]:
            raise NotImplementedError(
                f"Disabled '{cls.name}' backend: 'f{message}'\n",
                f"You can still enable the backend by hand using the environment variable '{enabled_env_var}=1'",
            )

        # Replace generate method with raise
        if not hasattr(cls, "generate"):
            raise ValueError(f"Coding error. Expected a generate method on {cls}")
        # Flag that it got disabled for register lookup
        cls.disabled = True  # type: ignore
        cls.generate = _no_generate  # type: ignore
        return cls

    return _decorator



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses
import functools
import inspect
import math
import operator
from builtins import bool, float, int, tuple  # noqa: A004 shadowing a Python built-in
from typing import Any, Callable, Final, Generic, ParamSpec, Tuple, TypeAlias, TypeVar, Union, cast

import numpy as np
from numpy import float32, float64, int8, int16, int32, int64, uint8, uint16, uint32, uint64

from gt4py._core import definitions as core_defs
from gt4py.next import common
from gt4py.next.common import Dimension, Field  # noqa: F401 [unused-import] for TYPE_BUILTINS
from gt4py.next.iterator import runtime
from gt4py.next.type_system import type_specifications as ts


PYTHON_TYPE_BUILTINS = [bool, int, float, tuple]
PYTHON_TYPE_BUILTIN_NAMES = [t.__name__ for t in PYTHON_TYPE_BUILTINS]

TYPE_BUILTINS = [
    common.Field,
    common.Dimension,
    int8,
    uint8,
    int16,
    uint16,
    int32,
    uint32,
    int64,
    uint64,
    float32,
    float64,
    *PYTHON_TYPE_BUILTINS,
]  # TODO(tehrengruber): validate matches iterator.builtins.TYPE_BUILTINS?

TYPE_BUILTIN_NAMES = [t.__name__ for t in TYPE_BUILTINS]

# Be aware: Type aliases are not fully supported in the frontend yet, e.g. `IndexType(1)` will not
# work.
IndexType: TypeAlias = int32

TYPE_ALIAS_NAMES = ["IndexType"]

_P = ParamSpec("_P")
_R = TypeVar("_R")


def _type_conversion_helper(t: type) -> type[ts.TypeSpec] | tuple[type[ts.TypeSpec], ...]:
    if t is common.Field:
        return ts.FieldType
    elif t is common.Dimension:
        return ts.DimensionType
    elif t is FieldOffset:
        return ts.OffsetType
    elif t is common.Connectivity:
        return ts.OffsetType
    elif t is core_defs.ScalarT:
        return ts.ScalarType
    elif t is common.Domain:
        return ts.DomainType
    elif t is type:
        return (
            ts.FunctionType
        )  # our type of type is currently represented by the type constructor function
    elif t is Tuple or (hasattr(t, "__origin__") and t.__origin__ is tuple):
        return ts.TupleType
    elif hasattr(t, "__origin__") and t.__origin__ is Union:
        types = [_type_conversion_helper(e) for e in t.__args__]  # type: ignore[attr-defined]
        assert all(type(t) is type and issubclass(t, ts.TypeSpec) for t in types)
        return cast(tuple[type[ts.TypeSpec], ...], tuple(types))  # `cast` to break the recursion
    else:
        raise AssertionError("Illegal type encountered.")


def _type_conversion(t: type) -> ts.DeferredType:
    return ts.DeferredType(constraint=_type_conversion_helper(t))


@dataclasses.dataclass(frozen=True)
class BuiltInFunction(Generic[_R, _P]):
    name: str = dataclasses.field(init=False)
    # `function` can be used to provide a default implementation for all `Field` implementations,
    # e.g. a fused multiply add could have a default implementation as a*b+c, but an optimized implementation for a specific `Field`
    function: Callable[_P, _R]

    def __post_init__(self) -> None:
        object.__setattr__(self, "name", f"{self.function.__module__}.{self.function.__name__}")
        object.__setattr__(self, "__doc__", self.function.__doc__)

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> _R:
        impl = self.dispatch(*args)
        return impl(*args, **kwargs)

    def dispatch(self, *args: Any) -> Callable[_P, _R]:
        arg_types = tuple(type(arg) for arg in args)
        for atype in arg_types:
            # current strategy is to select the implementation of the first arg that supports the operation
            # TODO: define a strategy that converts or prevents conversion
            if (dispatcher := getattr(atype, "__gt_builtin_func__", None)) is not None and (
                op_func := dispatcher(self)
            ) is not NotImplemented:
                return op_func
        else:
            return self.function

    def __gt_type__(self) -> ts.FunctionType:
        signature = inspect.signature(self.function)
        params = signature.parameters

        return ts.FunctionType(
            pos_only_args=[
                _type_conversion(param.annotation)
                for param in params.values()
                if param.kind == inspect.Parameter.POSITIONAL_ONLY
            ],
            pos_or_kw_args={
                k: _type_conversion(v.annotation)
                for k, v in params.items()
                if v.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
            },
            kw_only_args={
                k: _type_conversion(v.annotation)
                for k, v in params.items()
                if v.kind == inspect.Parameter.KEYWORD_ONLY
            },
            returns=_type_conversion(signature.return_annotation),
        )


CondT = TypeVar("CondT", bound=Union[common.Field, common.Domain])
FieldT = TypeVar("FieldT", bound=Union[common.Field, core_defs.Scalar, Tuple])


class WhereBuiltinFunction(
    BuiltInFunction[_R, [CondT, FieldT, FieldT]], Generic[_R, CondT, FieldT]
):
    def __call__(self, cond: CondT, true_field: FieldT, false_field: FieldT) -> _R:
        if isinstance(true_field, tuple) or isinstance(false_field, tuple):
            if not (isinstance(true_field, tuple) and isinstance(false_field, tuple)):
                raise ValueError(
                    # TODO(havogt) find a strategy to unify parsing and embedded error messages
                    f"Either both or none can be tuple in '{true_field=}' and '{false_field=}'."
                )
            if len(true_field) != len(false_field):
                raise ValueError(
                    "Tuple of different size not allowed."
                )  # TODO(havogt) find a strategy to unify parsing and embedded error messages
            return tuple(self(cond, t, f) for t, f in zip(true_field, false_field))  # type: ignore[return-value] # `tuple` is not `_R`
        return super().__call__(cond, true_field, false_field)


@BuiltInFunction
def neighbor_sum(field: common.Field, /, axis: common.Dimension) -> common.Field:
    raise NotImplementedError()


@BuiltInFunction
def max_over(field: common.Field, /, axis: common.Dimension) -> common.Field:
    raise NotImplementedError()


@BuiltInFunction
def min_over(field: common.Field, /, axis: common.Dimension) -> common.Field:
    raise NotImplementedError()


@BuiltInFunction
def broadcast(
    field: common.Field | core_defs.ScalarT, dims: tuple[common.Dimension, ...], /
) -> common.Field:
    assert core_defs.is_scalar_type(
        field
    )  # default implementation for scalars, Fields are handled via dispatch
    # TODO(havogt) implement with FunctionField, the workaround is to ignore broadcasting on scalars as they broadcast automatically, but we lose the check for compatible dimensions
    return field  # type: ignore[return-value] # see comment above


@WhereBuiltinFunction
def where(
    mask: common.Field,
    true_field: common.Field | core_defs.ScalarT | Tuple,
    false_field: common.Field | core_defs.ScalarT | Tuple,
    /,
) -> common.Field | Tuple:
    raise NotImplementedError()


@BuiltInFunction
def astype(
    value: common.Field | core_defs.ScalarT | Tuple, type_: type, /
) -> common.Field | core_defs.ScalarT | Tuple:
    if isinstance(value, tuple):
        return tuple(astype(v, type_) for v in value)
    # default implementation for scalars, Fields are handled via dispatch
    assert core_defs.is_scalar_type(value)
    return core_defs.dtype(type_).scalar_type(value)


_UNARY_MATH_NUMBER_BUILTIN_IMPL: Final = {"abs": abs, "neg": operator.neg}
UNARY_MATH_NUMBER_BUILTIN_NAMES: Final = [*_UNARY_MATH_NUMBER_BUILTIN_IMPL.keys()]

_UNARY_MATH_FP_BUILTIN_IMPL: Final = {
    "sin": math.sin,
    "cos": math.cos,
    "tan": math.tan,
    "arcsin": math.asin,
    "arccos": math.acos,
    "arctan": math.atan,
    "sinh": math.sinh,
    "cosh": math.cosh,
    "tanh": math.tanh,
    "arcsinh": math.asinh,
    "arccosh": math.acosh,
    "arctanh": math.atanh,
    "sqrt": math.sqrt,
    "exp": math.exp,
    "log": math.log,
    "gamma": math.gamma,
    "cbrt": math.cbrt if hasattr(math, "cbrt") else np.cbrt,  # match.cbrt() only added in 3.11
    "floor": math.floor,
    "ceil": math.ceil,
    "trunc": math.trunc,
}
UNARY_MATH_FP_BUILTIN_NAMES: Final = [*_UNARY_MATH_FP_BUILTIN_IMPL.keys()]

_UNARY_MATH_FP_PREDICATE_BUILTIN_IMPL: Final = {
    "isfinite": math.isfinite,
    "isinf": math.isinf,
    "isnan": math.isnan,
}
UNARY_MATH_FP_PREDICATE_BUILTIN_NAMES: Final = [*_UNARY_MATH_FP_PREDICATE_BUILTIN_IMPL.keys()]


def _make_unary_math_builtin(name: str) -> None:
    _math_builtin = (
        _UNARY_MATH_NUMBER_BUILTIN_IMPL
        | _UNARY_MATH_FP_BUILTIN_IMPL
        | _UNARY_MATH_FP_PREDICATE_BUILTIN_IMPL
    )[name]

    def impl(value: common.Field | core_defs.ScalarT, /) -> common.Field | core_defs.ScalarT:
        # TODO(havogt): enable tests in `test_math_builtin_execution.py`
        assert core_defs.is_scalar_type(
            value
        )  # default implementation for scalars, Fields are handled via dispatch

        return cast(common.Field | core_defs.ScalarT, _math_builtin(value))  # type: ignore[operator] # calling a function of unknown type

    impl.__name__ = name
    globals()[name] = BuiltInFunction(impl)


for f in (
    UNARY_MATH_NUMBER_BUILTIN_NAMES
    + UNARY_MATH_FP_BUILTIN_NAMES
    + UNARY_MATH_FP_PREDICATE_BUILTIN_NAMES
):
    _make_unary_math_builtin(f)

BINARY_MATH_NUMBER_BUILTIN_NAMES = ["minimum", "maximum", "fmod", "power"]


def _make_binary_math_builtin(name: str) -> None:
    def impl(
        lhs: common.Field | core_defs.ScalarT, rhs: common.Field | core_defs.ScalarT, /
    ) -> common.Field | core_defs.ScalarT:
        # default implementation for scalars, Fields are handled via dispatch
        assert core_defs.is_scalar_type(lhs)
        assert core_defs.is_scalar_type(rhs)
        return getattr(np, name)(lhs, rhs)

    impl.__name__ = name
    globals()[name] = BuiltInFunction(impl)


for f in BINARY_MATH_NUMBER_BUILTIN_NAMES:
    _make_binary_math_builtin(f)

MATH_BUILTIN_NAMES = (
    UNARY_MATH_NUMBER_BUILTIN_NAMES
    + UNARY_MATH_FP_BUILTIN_NAMES
    + UNARY_MATH_FP_PREDICATE_BUILTIN_NAMES
    + BINARY_MATH_NUMBER_BUILTIN_NAMES
)

FUN_BUILTIN_NAMES = [
    "neighbor_sum",
    "max_over",
    "min_over",
    "broadcast",
    "where",
    "astype",
    *MATH_BUILTIN_NAMES,
]

BUILTIN_NAMES = TYPE_BUILTIN_NAMES + FUN_BUILTIN_NAMES

BUILTINS = {name: globals()[name] for name in BUILTIN_NAMES}

__all__ = [*((set(BUILTIN_NAMES) | set(TYPE_ALIAS_NAMES)) - {"Dimension", "Field"})]


# TODO(tehrengruber): FieldOffset and runtime.Offset are not an exact conceptual
#  match. Revisit if we want to continue subclassing here. If we split
#  them also check whether Dimension should continue to be the shared or define
#  guidelines for decision.
@dataclasses.dataclass(frozen=True)
class FieldOffset(runtime.Offset):
    source: common.Dimension
    target: tuple[common.Dimension] | tuple[common.Dimension, common.Dimension]

    @functools.cached_property
    def _cache(self) -> dict:
        return {}

    def __post_init__(self) -> None:
        if len(self.target) == 2 and self.target[1].kind != common.DimensionKind.LOCAL:
            raise ValueError("Second dimension in offset must be a local dimension.")

    def __gt_type__(self) -> ts.OffsetType:
        return ts.OffsetType(source=self.source, target=self.target)

    def __getitem__(self, offset: int) -> common.Connectivity:
        """Serve as a connectivity factory."""
        from gt4py.next import embedded  # avoid circular import

        assert isinstance(self.value, str)
        current_offset_provider = embedded.context.get_offset_provider(None)
        assert current_offset_provider is not None
        offset_definition = current_offset_provider[self.value]

        connectivity: common.Connectivity
        if isinstance(offset_definition, common.Dimension):
            connectivity = common.CartesianConnectivity(offset_definition, offset)
        elif isinstance(offset_definition, common.Connectivity):
            assert common.is_neighbor_connectivity(offset_definition)
            named_index = common.NamedIndex(self.target[-1], offset)
            connectivity = offset_definition[named_index]
        else:
            raise NotImplementedError()

        return connectivity

    def as_connectivity_field(self) -> common.Connectivity:
        """Convert to connectivity field using the offset providers in current embedded execution context."""
        from gt4py.next import embedded  # avoid circular import

        assert isinstance(self.value, str)
        current_offset_provider = embedded.context.get_offset_provider(None)
        assert current_offset_provider is not None
        offset_definition = current_offset_provider[self.value]

        cache_key = id(offset_definition)
        if (connectivity := self._cache.get(cache_key, None)) is None:
            if isinstance(offset_definition, common.Connectivity):
                connectivity = offset_definition
            else:
                raise NotImplementedError()

            self._cache[cache_key] = connectivity

        return connectivity



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""
GT4Py-NEXT - Performance portable and composable weather & climate stencils.

This module deviates from the project coding style (Google Python style) in the following way:

[style guide link](https://google.github.io/styleguide/pyguide.html#22-imports):

Fnctions and classes are imported from other modules in order to explicitly re-export them
to create a streamlined user experience. `from module import *` can be used but only if the
module in question is a submodule, defines `__all__` and exports many public API objects.
"""

# ruff: noqa: F401
from . import common, ffront, iterator, program_processors
from .common import (
    Connectivity,
    Dimension,
    DimensionKind,
    Dims,
    Domain,
    Field,
    GridType,
    UnitRange,
    domain,
    unit_range,
)
from .constructors import as_connectivity, as_field, empty, full, ones, zeros
from .embedded import (  # Just for registering field implementations
    nd_array_field as _nd_array_field,
)
from .ffront import fbuiltins
from .ffront.decorator import field_operator, program, scan_operator
from .ffront.fbuiltins import *  # noqa: F403 [undefined-local-with-import-star]  explicitly reexport all from fbuiltins.__all__
from .ffront.fbuiltins import FieldOffset
from .iterator.embedded import (
    NeighborTableOffsetProvider,  # TODO(havogt): deprecated
    index_field,
    np_as_located_field,
)
from .program_processors.runners.gtfn import (
    run_gtfn_cached as gtfn_cpu,
    run_gtfn_gpu_cached as gtfn_gpu,
)
from .program_processors.runners.roundtrip import default as itir_python


__all__ = [
    # submodules
    "common",
    "ffront",
    "iterator",
    "program_processors",
    # from common
    "Dimension",
    "DimensionKind",
    "Field",
    "Connectivity",
    "GridType",
    "domain",
    "Domain",
    "unit_range",
    "UnitRange",
    # from constructors
    "empty",
    "zeros",
    "ones",
    "full",
    "as_field",
    "as_connectivity",
    # from iterator
    "NeighborTableOffsetProvider",
    "index_field",
    "np_as_located_field",
    # from ffront
    "FieldOffset",
    "field_operator",
    "program",
    "scan_operator",
    # from program_processor
    "gtfn_cpu",
    "gtfn_gpu",
    "itir_python",
    *fbuiltins.__all__,
]



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from collections.abc import Sequence
from typing import Any, Optional

import dace

from gt4py._core import definitions as core_defs
from gt4py.next import common as gtx_common, field_utils

from . import utils as gtx_dace_utils


def get_field_domain_symbols(name: str, domain: gtx_common.Domain) -> dict[str, int]:
    assert gtx_common.Domain.is_finite(domain)
    return {
        gtx_dace_utils.range_start_symbol(name, i): r.start for i, r in enumerate(domain.ranges)
    } | {gtx_dace_utils.range_stop_symbol(name, i): r.stop for i, r in enumerate(domain.ranges)}


def get_array_shape_symbols(
    array_desc: dace.data.Array, ndarray: core_defs.NDArrayObject
) -> dict[str, int]:
    array_symbols = {}
    for array_size, size in zip(array_desc.shape, ndarray.shape, strict=True):
        if (array_size == size) == True:  # noqa: E712 [true-false-comparison]  # SymPy Fancy comparison.
            pass
        else:
            assert isinstance(array_size, dace.symbol)
            array_symbols[array_size.name] = size
    return array_symbols


def get_array_stride_symbols(
    array_desc: dace.data.Array, ndarray: core_defs.NDArrayObject
) -> dict[str, int]:
    array_symbols = {}
    for array_stride, value in zip(array_desc.strides, ndarray.strides, strict=True):
        assert divmod(value, ndarray.itemsize)[1] == 0
        stride = value // ndarray.itemsize
        if (array_stride == stride) == True:  # noqa: E712 [true-false-comparison]  # SymPy Fancy comparison.
            pass
        else:
            assert isinstance(array_stride, dace.symbol)
            array_symbols[array_stride.name] = stride
    return array_symbols


def _convert_arg(arg: Any) -> tuple[Any, Optional[gtx_common.Domain]]:
    if (domain := getattr(arg, "domain", None)) is None:
        return arg, None
    if len(domain.dims) == 0:
        # Pass zero-dimensional fields as scalars.
        return arg.as_scalar(), None
    return arg.ndarray, domain


def _get_args(sdfg: dace.SDFG, args: Sequence[Any]) -> dict[str, Any]:
    call_args = {}
    range_symbols: dict[str, int] = {}
    stride_symbols: dict[str, int] = {}
    for name, arg in zip(sdfg.arg_names, args, strict=True):
        call_arg, domain = _convert_arg(arg)
        call_args[name] = call_arg
        if domain is not None:
            range_symbols |= get_field_domain_symbols(name, domain)
            stride_symbols |= get_array_stride_symbols(sdfg.arrays[name], call_arg)
    # sanity check in case range symbols are passed as explicit program arguments
    assert all(
        call_arg == range_symbols[param]
        for param, call_arg in call_args.items()
        if param in range_symbols
    )
    return call_args | range_symbols | stride_symbols


def get_sdfg_conn_args(
    sdfg: dace.SDFG,
    offset_provider: gtx_common.OffsetProvider,
) -> dict[str, core_defs.NDArrayObject]:
    """
    Extracts the connectivity tables that are used in the sdfg and ensures
    that the memory buffers are allocated for the target device.
    """
    connectivity_args = {}
    for offset, connectivity in offset_provider.items():
        name = gtx_dace_utils.connectivity_identifier(offset)
        if name in sdfg.arrays:
            assert gtx_common.is_neighbor_connectivity(connectivity)
            assert field_utils.verify_device_field_type(
                connectivity,
                core_defs.CUPY_DEVICE_TYPE
                if sdfg.arrays[name].storage == dace.dtypes.StorageType.GPU_Global
                else core_defs.DeviceType.CPU,  # type: ignore[arg-type] # if we are `on_gpu` we expect `CUPY_DEVICE_TYPE` to be not `None`
            )
            connectivity_args[name] = connectivity.ndarray

    return connectivity_args


def get_sdfg_args(
    sdfg: dace.SDFG,
    offset_provider: gtx_common.OffsetProvider,
    *args: Any,
    filter_args: bool = False,
) -> dict[str, Any]:
    """Extracts the arguments needed to call the SDFG.

    This function can handle the arguments that are passed to the dace runner
    and that end up in the decoration stage of the dace backend workflow.

    Args:
        sdfg:               The SDFG for which we want to get the arguments.
        offset_provider:    The offset provider.
        args:               The list of arguments passed to the dace runner.
        filter_args:        If True, return only the arguments that are expected
                            according to the SDFG signature.

    Returns:
        A dictionary of keyword arguments to be passed in the SDFG call.
    """

    call_args = _get_args(sdfg, args)
    connectivity_args = get_sdfg_conn_args(sdfg, offset_provider)
    for conn, ndarray in connectivity_args.items():
        call_args |= get_array_shape_symbols(sdfg.arrays[conn], ndarray)
        call_args |= get_array_stride_symbols(sdfg.arrays[conn], ndarray)
    call_args |= connectivity_args

    if filter_args:
        # return only arguments expected in SDFG signature (note hat `signature_arglist` takes time)
        sdfg_sig = sdfg.signature_arglist(with_types=False)
        return {key: call_args[key] for key in sdfg_sig}

    return call_args
